from __future__ import annotations

import json

from palimpzest.core.data.dataclasses import PlanCost


def construct_policy_from_kwargs(**kwargs) -> Policy | None:
    """
    Construct and return a policy object which is defined by the keyword arguments.

    This function accepts the following keyword arguments:
    - max_quality
    - min_cost
    - min_time
    - cost_budget
    - time_budget
    - quality_threshold

    If none of these keyword arguments are provided, the function will return None.
    """
    # compute the number of objectives and constraints in the kwargs
    num_objectives = sum([bool(kwargs.get(key, False)) for key in ["max_quality", "min_cost", "min_time"]])
    num_constraints = sum([bool(kwargs.get(key, False)) for key in ["cost_budget", "time_budget", "quality_threshold"]])

    # if there are no policy kwargs provided, return None
    if num_objectives == 0 and num_constraints == 0:
        return None

    # Otherwise, assert that kwargs are valid
    assert num_objectives == 1, "Must optimize for one of max_quality, min_cost, or min_time."
    assert num_constraints <= 1, "Currently, PZ supports at most one constraint."

    # print warning if user tries to set a constraint and optimization goal on the same metric
    if "max_quality" in kwargs and "quality_threshold" in kwargs:
        print("Warning: Setting a constraint on quality and optimizing for quality is redundant.")

    if "min_cost" in kwargs and "cost_budget" in kwargs:
        print("Warning: Setting a constraint on cost and optimizing for cost is redundant.")

    if "min_time" in kwargs and "time_budget" in kwargs:
        print("Warning: Setting a constraint on time and optimizing for time is redundant.")

    # construct the policy object
    policy = None
    if "max_quality" in kwargs and "cost_budget" in kwargs:
        policy = MaxQualityAtFixedCost(kwargs["cost_budget"])
    elif "max_quality" in kwargs and "time_budget" in kwargs:
        policy = MaxQualityAtFixedTime(kwargs["time_budget"])
    elif "max_quality" in kwargs:
        policy = MaxQuality()
    elif "min_cost" in kwargs and "quality_threshold" in kwargs:
        policy = MinCostAtFixedQuality(kwargs["quality_threshold"])
    elif "min_cost" in kwargs:
        policy = MinCost()
    elif "min_time" in kwargs and "quality_threshold" in kwargs:
        policy = MinTimeAtFixedQuality(kwargs["quality_threshold"])
    elif "min_time" in kwargs:
        policy = MinTime()

    return policy


class Policy:
    """
    Base class for a policy. Each policy has two methods: constraint() and chooose().
    The first method determines whether the given cost, runtime, and quality for a plan
    (or sub-plan) satisfy the policy's constraint(s). The second method takes in the 
    (cost, runtime, quality) tuples for two plans (or subplans) and returns True if the
    first plan is better than the second one and False otherwise.
    """

    def __init__(self):
        pass

    def get_primary_metric(self) -> str:
        """
        Returns one of ["cost", "time", "quality"]; whichever corresponds to the
        maximization / minimization goal of the policy.

        Eventually we may make policies more general by allowing users to optimize
        some function: f(cost, time, quality). In that case, we may deprecate this
        method and update its callers.
        """
        raise NotImplementedError("Calling this method from an abstract base class.")

    def get_dict(self) -> dict:
        """
        Returns a dict representation of the policy which specifies how much weight
        (in [0,1]) should be given to each metric.
        """
        raise NotImplementedError("Calling this method from an abstract base class.")

    def constraint(self, plan: PlanCost) -> bool:
        """
        Return True if the given (cost, runtime, quality) for a plan (or subplan)
        satisfy the policy's constraint(s). Otherwise, return False.
        """
        raise NotImplementedError("Calling this method from an abstract base class.")

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan is better than other_plan and return False otherwise.
        """
        raise NotImplementedError("Calling this method from an abstract base class.")

    def to_json_str(self) -> str:
        """Convert policy configuration to a JSON-serializable dictionary."""
        return json.dumps({
            "type": self.__class__.__name__,
            "config": self.get_dict()
        }, indent=2)

class MaxQuality(Policy):
    """
    This policy has no constraints and computes the best plan as the one with
    the higher quality.
    """

    def __str__(self):
        return "Maximum Quality"

    def get_primary_metric(self) -> str:
        return "quality"

    def get_dict(self) -> dict:
        return {"cost": 0.0, "time": 0.0, "quality": 1.0}

    def constraint(self, plan: PlanCost) -> bool:
        """There is no constraint."""
        return True

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has higher quality than other_plan and return False otherwise.
        Use cost and then runtime as tiebreakers.
        """
        if plan.quality == other_plan.quality:
            if plan.cost == other_plan.cost:
                return plan.time < other_plan.time
            return plan.cost < other_plan.cost

        return plan.quality > other_plan.quality


class MinCost(Policy):
    """
    This policy has no constraints and computes the best plan as the one with
    the lower cost.
    """

    def __str__(self):
        return "Minimum Cost"

    def get_primary_metric(self) -> str:
        return "cost"

    def get_dict(self) -> dict:
        return {"cost": 1.0, "time": 0.0, "quality": 0.0}

    def constraint(self, plan: PlanCost) -> bool:
        """There is no constraint."""
        return True

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has lower cost than other_plan and return False otherwise.
        Use quality and then runtime as tiebreakers.
        """
        if plan.cost == other_plan.cost:
            if plan.quality == other_plan.quality:
                return plan.time < other_plan.time
            return plan.quality > other_plan.quality

        return plan.cost < other_plan.cost


class MinTime(Policy):
    """
    This policy has no constraints and computes the best plan as the one with
    the lower runtime.
    """

    def __str__(self):
        return "Minimum Time"

    def get_primary_metric(self) -> str:
        return "time"

    def get_dict(self) -> dict:
        return {"cost": 0.0, "time": 1.0, "quality": 0.0}

    def constraint(self, plan: PlanCost) -> bool:
        """There is no constraint."""
        return True

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has lower runtime than other_plan and return False otherwise.
        Use quality and then cost as tiebreakers.
        """
        if plan.time == other_plan.time:
            if plan.quality == other_plan.quality:
                return plan.cost < other_plan.cost
            return plan.quality > other_plan.quality

        return plan.time < other_plan.time


class MaxQualityAtFixedCost(Policy):
    """
    This policy applies a constraint (upper bound) on the cost of the plan
    and tries to maximize quality subject to that constraint.
    """

    def __init__(self, max_cost: float):
        self.max_cost = max_cost

    def __str__(self):
        return "MaxQuality@FixedCost"

    def get_primary_metric(self) -> str:
        return "quality"

    def get_dict(self) -> dict:
        return {"cost": 0.5, "time": 0.0, "quality": 0.5}

    def constraint(self, plan: PlanCost) -> bool:
        return plan.cost < self.max_cost

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has higher quality than other_plan and return False otherwise.
        Use cost and then runtime as a tie-breaker.
        """
        if plan.quality == other_plan.quality:
            if plan.cost == other_plan.cost:
                return plan.time < other_plan.time
            return plan.cost < other_plan.cost

        return plan.quality > other_plan.quality


class MaxQualityAtFixedTime(Policy):
    """
    This policy applies a constraint (upper bound) on the runtime of the plan
    and tries to maximize quality subject to that constraint.
    """

    def __init__(self, max_time: float):
        self.max_time = max_time

    def __str__(self):
        return "MaxQuality@FixedTime"

    def get_primary_metric(self) -> str:
        return "quality"

    def get_dict(self) -> dict:
        return {"cost": 0.0, "time": 0.5, "quality": 0.5}

    def constraint(self, plan: PlanCost) -> bool:
        return plan.time < self.max_time

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has higher quality than other_plan and return False otherwise.
        Use runtime and then cost as a tie-breaker.
        """
        if plan.quality == other_plan.quality:
            if plan.time == other_plan.time:
                return plan.cost < other_plan.cost
            return plan.time < other_plan.time

        return plan.quality > other_plan.quality


class MinCostAtFixedQuality(Policy):
    """
    This policy applies a constraint (lower bound) on the quality of the plan
    and tries to minimize cost subject to that constraint.
    """

    def __init__(self, min_quality: float):
        self.min_quality = min_quality

    def __str__(self):
        return "MinCost@FixedQuality"

    def get_primary_metric(self) -> str:
        return "cost"

    def get_dict(self) -> dict:
        return {"cost": 0.5, "time": 0.0, "quality": 0.5}

    def constraint(self, plan: PlanCost) -> bool:
        return plan.quality > self.min_quality

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has lower cost than other_plan and return False otherwise.
        Use quality and then runtime as a tie-breaker.
        """
        if plan.cost == other_plan.cost:
            if plan.quality == other_plan.quality:
                return plan.time < other_plan.time
            return plan.quality > other_plan.quality

        return plan.cost < other_plan.cost


class MinTimeAtFixedQuality(Policy):
    """
    This policy applies a constraint (lower bound) on the quality of the plan
    and tries to minimize runtime subject to that constraint.
    """

    def __init__(self, min_quality: float):
        self.min_quality = min_quality

    def __str__(self):
        return "MinTime@FixedQuality"

    def get_primary_metric(self) -> str:
        return "time"

    def get_dict(self) -> dict:
        return {"cost": 0.0, "time": 0.5, "quality": 0.5}

    def constraint(self, plan: PlanCost) -> bool:
        return plan.quality > self.min_quality

    def choose(self, plan: PlanCost, other_plan: PlanCost) -> float:
        """
        Return True if plan has lower runtime than other_plan and return False otherwise.
        Use quality and then cost as a tie-breaker.
        """
        if plan.time == other_plan.time:
            if plan.quality == other_plan.quality:
                return plan.cost < other_plan.cost
            return plan.quality > other_plan.quality

        return plan.time < other_plan.time


# TODO: add this back in a way which allows users to select a plan from a small pareto optimal set at the end of
# query optimization
# class UserChoice(Policy):
#     """
#     This policy asks the user to decide which of the pareto-optimal
#     candidate plans to execute.
#     """

#     def __str__(self):
#         return "User Choice"

#     def choose(self, candidatePlans: List[PhysicalPlan]) -> PhysicalPlan:
#         print("Please select a plan from the following options:")
#         for idx, plan in enumerate(candidatePlans):
#             print(f"[{idx}] {plan}")
#         user_choice = input(f"Please select a plan in [0-{len(candidatePlans) - 1}]: ")
#         user_choice = int(user_choice)
#         if user_choice not in range(len(candidatePlans)):
#             print(
#                 f"Error: user choice {user_choice} was not a number in the specified range. Please try again."
#             )
#             return self.choose(candidatePlans)

#         return candidatePlans[user_choice]



import logging
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, wait

import numpy as np
from chromadb.api.models.Collection import Collection

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS
from palimpzest.core.data.dataclasses import OperatorCostEstimates, PlanStats, RecordOpStats
from palimpzest.core.data.datareaders import DataReader
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.policy import Policy
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import FilterOp, LLMFilter
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.retrieve import RetrieveOp
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan
from palimpzest.utils.progress import PZSentinelProgressManager

logger = logging.getLogger(__name__)

class BaseExecutionStrategy:
    def __init__(self,
                 scan_start_idx: int = 0, 
                 max_workers: int | None = None,
                 num_samples: int | None = None,
                 cache: bool = False,
                 verbose: bool = False,
                 progress: bool = True,
                 *args,
                 **kwargs):
        self.scan_start_idx = scan_start_idx
        self.max_workers = max_workers
        self.num_samples = num_samples
        self.cache = cache
        self.verbose = verbose
        self.progress = progress


    def _add_records_to_cache(self, target_cache_id: str, records: list[DataRecord]) -> None:
        """Add each record (which isn't filtered) to the cache for the given target_cache_id."""
        if self.cache:
            for record in records:
                if getattr(record, "passed_operator", True):
                    # self.datadir.append_cache(target_cache_id, record)
                    pass

    def _close_cache(self, target_cache_ids: list[str]) -> None:
        """Close the cache for each of the given target_cache_ids"""
        if self.cache:
            for target_cache_id in target_cache_ids:  # noqa: B007
                # self.datadir.close_cache(target_cache_id)
                pass

class ExecutionStrategy(BaseExecutionStrategy, ABC):
    """Base strategy for executing query plans. Defines how to execute a PhysicalPlan.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logger.info(f"Initialized ExecutionStrategy {self.__class__.__name__}")
        logger.debug(f"ExecutionStrategy initialized with config: {self.__dict__}")

    @abstractmethod
    def execute_plan(self, plan: PhysicalPlan) -> tuple[list[DataRecord], PlanStats]:
        """Execute a single plan according to strategy"""
        pass

    def _create_input_queues(self, plan: PhysicalPlan) -> dict[str, list]:
        """Initialize input queues for each operator in the plan."""
        input_queues = {}
        for op in plan.operators:
            inputs = []
            if isinstance(op, ScanPhysicalOp):
                scan_end_idx = (
                    len(op.datareader)
                    if self.num_samples is None
                    else min(self.scan_start_idx + self.num_samples, len(op.datareader))
                )
                inputs = [idx for idx in range(self.scan_start_idx, scan_end_idx)]
            input_queues[op.get_full_op_id()] = inputs

        return input_queues

class SentinelExecutionStrategy(BaseExecutionStrategy, ABC):
    """Base strategy for executing sentinel query plans. Defines how to execute a SentinelPlan."""
    """
    Specialized query processor that implements MAB sentinel strategy
    for coordinating optimization and execution.
    """
    def __init__(
        self,
        val_datasource: DataReader,
        k: int,
        j: int,
        sample_budget: int,
        policy: Policy,
        priors: dict | None = None,
        use_final_op_quality: bool = False,
        seed: int = 42,
        exp_name: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.val_datasource = val_datasource
        self.k = k
        self.j = j
        self.sample_budget = sample_budget
        self.policy = policy
        self.priors = priors
        self.use_final_op_quality = use_final_op_quality
        self.seed = seed
        self.rng = np.random.default_rng(seed=seed)
        self.exp_name = exp_name

        # special cache which is used for tracking the target record sets for each (source_idx, logical_op_id)
        self.champion_output_cache: dict[int, dict[str, tuple[DataRecordSet, float]]] = {}

        # general cache which maps hash(logical_op_id, phys_op_id, hash(input)) --> record_set
        self.cache: dict[int, DataRecordSet] = {}

        # progress manager used to track progress of the execution
        self.progress_manager: PZSentinelProgressManager | None = None

    def _compute_quality(
            self,
            physical_op_cls: type[PhysicalOperator],
            record_set: DataRecordSet,
            target_record_set: DataRecordSet,
        ) -> DataRecordSet:
        """
        Compute the quality for the given `record_set` by comparing it to the `target_record_set`.

        Update the record_set by assigning the quality to each entry in its record_op_stats and
        returning the updated record_set.
        """
        # if this operation failed
        if len(record_set) == 0:
            record_set.record_op_stats[0].quality = 0.0

        # if this operation is a filter:
        # - return 1.0 if there's a match in the expected output which this operator does not filter out and 0.0 otherwise
        elif issubclass(physical_op_cls, FilterOp):
            # NOTE: we know that record_set.data_records will contain a single entry for a filter op
            record = record_set.data_records[0]

            # search for a record in the target with the same set of fields
            found_match_in_target = False
            for target_record in target_record_set:
                all_correct = True
                for field, value in record.field_values.items():
                    if value != target_record[field]:
                        all_correct = False
                        break

                if all_correct:
                    found_match_in_target = target_record.passed_operator
                    break

            # set quality based on whether we found a match in the target and return
            record_set.record_op_stats[0].quality = int(record.passed_operator == found_match_in_target)

            return record_set

        # if this is a successful convert operation
        else:
            # NOTE: the following computation assumes we do not project out computed values
            #       (and that the validation examples provide all computed fields); even if
            #       a user program does add projection, we can ignore the projection on the
            #       validation dataset and use the champion model (as opposed to the validation
            #       output) for scoring fields which have their values projected out

            # GREEDY ALGORITHM
            # for each record in the expected output, we look for the computed record which maximizes the quality metric;
            # once we've identified that computed record we remove it from consideration for the next expected output
            field_to_score_fn = target_record_set.get_field_to_score_fn()
            for target_record in target_record_set:
                best_quality, best_record_op_stats = 0.0, None
                for record_op_stats in record_set.record_op_stats:
                    # if we already assigned this record a quality, skip it
                    if record_op_stats.quality is not None:
                        continue

                    # compute number of matches between this record's computed fields and this expected record's outputs
                    total_quality = 0
                    for field in record_op_stats.generated_fields:
                        computed_value = record_op_stats.record_state.get(field, None)
                        expected_value = target_record[field]

                        # get the metric function for this field
                        score_fn = field_to_score_fn.get(field, "exact")

                        # compute exact match
                        if score_fn == "exact":
                            total_quality += int(computed_value == expected_value)

                        # compute UDF metric
                        elif callable(score_fn):
                            total_quality += score_fn(computed_value, expected_value)

                        # otherwise, throw an exception
                        else:
                            raise Exception(f"Unrecognized score_fn: {score_fn}")

                    # compute recall and update best seen so far
                    quality = total_quality / len(record_op_stats.generated_fields)
                    if quality > best_quality:
                        best_quality = quality
                        best_record_op_stats = record_op_stats

                # set best_quality as quality for the best_record_op_stats
                if best_record_op_stats is not None:
                    best_record_op_stats.quality = best_quality

        # for any records which did not receive a quality, set it to 0.0 as these are unexpected extras
        for record_op_stats in record_set.record_op_stats:
            if record_op_stats.quality is None:
                record_op_stats.quality = 0.0

        return record_set

    def _score_quality(
        self,
        physical_op_cls: type[PhysicalOperator],
        source_idx_to_record_sets: dict[int, list[DataRecordSet]],
        source_idx_to_target_record_set: dict[int, DataRecordSet],
    ) -> dict[int, list[DataRecordSet]]:
        """
        NOTE: This approach to cost modeling does not work directly for aggregation queries;
              for these queries, we would ask the user to provide validation data for the step immediately
              before a final aggregation

        NOTE: This function currently assumes that one-to-many converts do NOT create duplicate outputs.
        This assumption would break if, for example, we extracted the breed of every dog in an image.
        If there were two golden retrievers and a bernoodle in an image and we extracted:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}

        This function would currently give perfect accuracy to the following output:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}

        Even though it is missing one of the golden retrievers.
        """
        # extract information about the logical operation performed at this stage of the sentinel plan;
        # NOTE: we can infer these fields from context clues, but in the long-term we should have a more
        #       principled way of getting these directly from attributes either stored in the sentinel_plan
        #       or in the PhysicalOperator
        is_perfect_quality_op = (
            not issubclass(physical_op_cls, LLMConvert)
            and not issubclass(physical_op_cls, LLMFilter)
            and not issubclass(physical_op_cls, RetrieveOp)
        )

        # compute quality of each output computed by this operator
        for source_idx, record_sets in source_idx_to_record_sets.items():
            # if this operation does not involve an LLM, every record_op_stats object gets perfect quality
            if is_perfect_quality_op:
                for record_set in record_sets:
                    for record_op_stats in record_set.record_op_stats:
                        record_op_stats.quality = 1.0
                continue

            # extract target output for this record set
            target_record_set = source_idx_to_target_record_set[source_idx]

            # for each record_set produced by an operation, compute its quality
            for record_set in record_sets:
                record_set = self._compute_quality(physical_op_cls, record_set, target_record_set)

        # return the quality annotated record sets
        return source_idx_to_record_sets

    def _get_target_record_sets(
        self,
        logical_op_id: str,
        source_idx_to_record_set_tuples: dict[int, list[tuple[DataRecordSet, PhysicalOperator, bool]]],
        expected_outputs: dict[int, dict] | None,
    ) -> dict[int, DataRecordSet]:
        # initialize mapping from source index to target record sets
        source_idx_to_target_record_set = {}

        for source_idx, record_set_tuples in source_idx_to_record_set_tuples.items():
            # get the first generated output for this source_idx
            base_target_record = None
            for record_set, _, _ in record_set_tuples:
                if len(record_set) > 0:
                    base_target_record = record_set[0]
                    break

            # compute availability of data
            base_target_present = base_target_record is not None
            labels_present = expected_outputs is not None
            labels_for_source_present = False
            if labels_present and source_idx in expected_outputs:
                labels = expected_outputs[source_idx].get("labels", [])
                labels_dict_lst = labels if isinstance(labels, list) else [labels]
                labels_for_source_present = labels_dict_lst != [] and labels_dict_lst != [None]

            # if we have a base target record and label info, use the label info to construct the target record set
            if base_target_present and labels_for_source_present:
                # get the field_to_score_fn                
                field_to_score_fn = expected_outputs[source_idx].get("score_fn", {})

                # construct the target record set; we force passed_operator to be True for all target records
                target_records = []
                for labels_dict in labels_dict_lst:
                    target_record = base_target_record.copy()
                    for field, value in labels_dict.items():
                        target_record[field] = value
                    target_record.passed_operator = True
                    target_records.append(target_record)

                source_idx_to_target_record_set[source_idx] = DataRecordSet(target_records, None, field_to_score_fn)
                continue

            # get the best computed output for this (source_idx, logical_op_id) so far (if one exists)
            champion_record_set, champion_op_quality = None, None
            if source_idx in self.champion_output_cache and logical_op_id in self.champion_output_cache[source_idx]:
                champion_record_set, champion_op_quality = self.champion_output_cache[source_idx][logical_op_id]

            # get the highest quality output that we just computed
            max_quality_record_set, max_op_quality = self._pick_champion_output(record_set_tuples)

            # if this new output is of higher quality than our previous champion (or if we didn't have
            # a previous champion) then we update our champion record set
            if champion_op_quality is None or (max_op_quality is not None and max_op_quality > champion_op_quality):
                champion_record_set, champion_op_quality = max_quality_record_set, max_op_quality

            # update the cache with the new champion record set and quality
            if source_idx not in self.champion_output_cache:
                self.champion_output_cache[source_idx] = {}
            self.champion_output_cache[source_idx][logical_op_id] = (champion_record_set, champion_op_quality)

            # set the target
            source_idx_to_target_record_set[source_idx] = champion_record_set

        return source_idx_to_target_record_set

    def _pick_champion_output(self, record_set_tuples: list[tuple[DataRecordSet, PhysicalOperator, bool]]) -> tuple[DataRecordSet, float | None]:
        # find the operator with the highest estimated quality and return its record_set
        base_op_cost_est = OperatorCostEstimates(cardinality=1.0, cost_per_record=0.0, time_per_record=0.0, quality=1.0)
        champion_record_set, champion_quality = None, None
        for record_set, op, _ in record_set_tuples:
            # skip failed operations
            if len(record_set) == 0:
                continue

            # get the estimated quality of this operator
            est_quality = op.naive_cost_estimates(base_op_cost_est).quality if self._is_llm_op(op) else 1.0
            if champion_quality is None or est_quality > champion_quality:
                champion_record_set, champion_quality = record_set, est_quality

        return champion_record_set, champion_quality

    def _flatten_record_sets(self, source_idx_to_record_sets: dict[int, list[DataRecordSet]]) -> tuple[list[DataRecord], list[RecordOpStats]]:
        """
        Flatten the list of record sets and record op stats for each source_idx.
        """
        all_records, all_record_op_stats = [], []
        for _, record_sets in source_idx_to_record_sets.items():
            for record_set in record_sets:
                all_records.extend(record_set.data_records)
                all_record_op_stats.extend(record_set.record_op_stats)

        return all_records, all_record_op_stats

    def _execute_op_set(self, op_input_pairs: list[tuple[PhysicalOperator, DataRecord | int]]) -> tuple[dict[int, list[tuple[DataRecordSet, PhysicalOperator, bool]]], dict[str, int]]:
        def execute_op_wrapper(operator, input) -> tuple[DataRecordSet, PhysicalOperator, DataRecord | int]:
            record_set = operator(input)
            return record_set, operator, input

        # TODO: modify unit tests to always have record_op_stats so we can use record_op_stats for source_idx
        # for scan operators, `input` will be the source_idx
        def get_source_idx(input):
            return input.source_idx if isinstance(input, DataRecord) else input

        def get_hash(operator, input):
            return hash(f"{operator.get_full_op_id()}{hash(input)}")

        # initialize mapping from source indices to output record sets
        source_idx_to_record_sets_and_ops = {get_source_idx(input): [] for _, input in op_input_pairs}

        # if any operations were previously executed, read the results from the cache
        final_op_input_pairs = []
        for operator, input in op_input_pairs:
            # compute hash
            op_input_hash = get_hash(operator, input)

            # get result from cache
            if op_input_hash in self.cache:
                source_idx = get_source_idx(input)
                record_set, operator = self.cache[op_input_hash]
                source_idx_to_record_sets_and_ops[source_idx].append((record_set, operator, False))

            # otherwise, add to final_op_input_pairs
            else:
                final_op_input_pairs.append((operator, input))

        # keep track of the number of llm operations
        num_llm_ops = 0

        # create thread pool w/max workers and run futures over worker pool
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # create futures
            futures = [
                executor.submit(execute_op_wrapper, operator, input)
                for operator, input in final_op_input_pairs
            ]
            output_record_sets = []
            while len(futures) > 0:
                done_futures, not_done_futures = wait(futures, timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)
                for future in done_futures:
                    # update output record sets
                    record_set, operator, input = future.result()
                    output_record_sets.append((record_set, operator, input))

                    # update cache
                    op_input_hash = get_hash(operator, input)
                    self.cache[op_input_hash] = (record_set, operator)

                    # update progress manager
                    if self._is_llm_op(operator):
                        num_llm_ops += 1
                        self.progress_manager.incr(operator.get_logical_op_id(), num_samples=1, total_cost=record_set.get_total_cost())

                # update futures
                futures = list(not_done_futures)

            # update mapping from source_idx to record sets and operators
            for record_set, operator, input in output_record_sets:
                # get the source_idx associated with this input record;
                source_idx = get_source_idx(input)

                # add record_set to mapping from source_idx --> record_sets
                source_idx_to_record_sets_and_ops[source_idx].append((record_set, operator, True))

        return source_idx_to_record_sets_and_ops, num_llm_ops

    def _is_llm_op(self, physical_op: PhysicalOperator) -> bool:
        is_llm_convert = isinstance(physical_op, LLMConvert)
        is_llm_filter = isinstance(physical_op, LLMFilter)
        is_llm_retrieve = isinstance(physical_op, RetrieveOp) and isinstance(physical_op.index, Collection)
        return is_llm_convert or is_llm_filter or is_llm_retrieve

    @abstractmethod
    def execute_sentinel_plan(self, sentinel_plan: SentinelPlan, expected_outputs: dict[str, dict]):
        """Execute a SentinelPlan according to strategy"""
        pass



"""
This module is responsible for building schemas dynamically, taking an input file and generating a schema for it.
The class offers a general class method, from_file, that takes a file and generates a schema for it.
This method is a simple wrapper for different methods, e.g., from_csv, from_yml, etc.

"""

import json
import os

import pandas as pd
import yaml
from pyld import jsonld

import palimpzest.core.lib.fields as pz_fields
import palimpzest.core.lib.schemas as pz_schemas
from palimpzest.core.lib.fields import Field
from palimpzest.core.lib.schemas import Schema


class SchemaBuilder:

    @classmethod
    def from_file(cls,
        schema_file: str,
        schema_name: str = "",
        schema_description: str = "",
        include_attributes: list = None,
        exclude_attributes: list = None,
        schema_type: Schema = None,
        ):
        """
        Inputs:
            schema_file: str - the path to the file
            description (optional): str - the description of the schema
            name (optional): str - the name of the schema
            include_attributes (optional): list - a list of attribute names to include in the schema. If None, all attributes are included.
            exclude_attributes (optional): list - a list of attribute names to exclude from the schema. If None, no attributes are excluded.
            schema_type (optional): Schema - the parent type of the schema to generate, e.g. ScientificPapers have a schema_type of PDFFile. If None, a generic Schema type is used.
        Outputs:
            A class object - the dynamically generated class
        """

        # Get the file extension
        filename = os.path.basename(schema_file)
        basename, file_extension = os.path.splitext(filename)
      
        if file_extension == ".csv":
            schema_data = cls.from_csv(schema_file)
        elif file_extension == ".json":
            schema_data = cls.from_json(schema_file)
        elif file_extension == ".jsonld":
            schema_data = cls.from_jsonld(schema_file)
        elif file_extension == ".yml":
            schema_data = cls.from_yml(schema_file)
        else:
            raise ValueError("Unsupported file format!")

        # If additional metadata is not provided, read it from the file. 
        # If not available, generate it from the filename
        if not schema_name:
            if schema_data['name']:
                schema_name = schema_data['name']
            else:
                schema_name = "".join([word.capitalize() for word in basename.split("_")])

        if not schema_description:
            if schema_data['description']:
                schema_description = schema_data['description']
            else:
                schema_description = f"A schema generated from the {file_extension} file {basename}."

        if include_attributes is None:
           include_attributes = []
        if exclude_attributes is None:
           exclude_attributes = []

        if schema_type is None:
            if schema_data.get('type', None):
                # Find if the schema type is a valid class in pz
                parsed_type = getattr(pz_schemas
                                      , schema_data['type'], Schema)
                schema_type = parsed_type if issubclass(parsed_type, Schema) else Schema
            else:
                schema_type = Schema

        # Generate the schema class dynamically
        attributes = {"__doc__": schema_description}
        include_attributes_lower = set([a.lower() for a in include_attributes])
        exclude_attributes_lower = set([a.lower() for a in exclude_attributes])
        for field in schema_data['fields']:
            norm_name = field['name'].lower()
            if len(include_attributes) and norm_name not in include_attributes_lower:
                continue
            if norm_name in exclude_attributes_lower:
                continue
            name = field['name']
            description = field.get('description', '')
            field_type = field.get('type', 'Field')
            field_type = getattr(pz_fields, field_type, Field)
            if not issubclass(field_type, Field):
                field_type = Field
                  
            attributes[name] = field_type(desc=description)

        # Create the class dynamically
        return type(schema_name, (schema_type,), attributes)

    @classmethod
    def from_csv(
        cls,
        schema_file: str,
    ) -> dict:
        """
        The attributes are extracted from the column names of the CSV file.
        If columns contain null values, they are marked as optional.
        TODO: Find a way to infer the description of the field.
        """
        
        # Use pandas to read the CSV file
        df = pd.read_csv(schema_file)
        columns = df.columns.tolist()

        # Generate the schema class dynamically
        fields = []
        for col in columns:
            field_type = df[col].dtype
            if field_type == float or field_type == int:  # noqa
                field_type = "NumericField"
            else:
                field_type = "Field"

            fields.append({"name":col,
                           "description":"",
                           "type":field_type})
        
        return {
            "name": '',
            "description": '',
            "fields": fields,
        }

    @classmethod
    def from_jsonld(
        cls,
        schema_file: str,
    ) -> dict:
        """JSON-LD schema builder.
        The attributes are extracted from the JSON-LD objects of type 'rdfs:Class'.
        If they contain a 'comment' field, this is used to populate a schema descripton.
        If they contain a 'rangeIncludes' field, this is used within the description to
        signal the list of valid values.
       """

        # Load the schema from the JSONLD file
        with open(schema_file) as file:
            jsonld_data = json.load(file)
        context = jsonld_data.get("@context")
        compacted_data = jsonld.compact(jsonld_data, context)
        compacted_graph = compacted_data["@graph"]

        fields = []

        for node in compacted_graph:
            if node.get("@type") != "rdfs:Class":
                continue
            name = node.get("rdfs:label")

            values = []
            if "schema:rangeIncludes" in node:
                values = [val["@id"].split(":")[-1] for val in node["schema:rangeIncludes"]]
            
            description = node.get("rdfs:comment", "")
            if values:
                description += " The only valid values are: " + ", ".join(values)
            fields.append({
                "name": name,
                "description": description, 
                "values": values})

        return {
            "name": '',
            "description": '',
            "fields": fields,
        }
    
    @classmethod
    def from_json(
            cls,
            schema_file: str,
            schema_name: str = "",
            schema_description: str = "",
            include_attributes: list = None,
            schema_type: Schema = None,
    )-> dict:
        """
        The attributes are extracted from the JSON objects.
        The format of the json has to be in the form:
        {
            "attribute1": {
                "description": "description",
            },
            ...
        }
        """

        # Load the schema from the JSON file
        with open(schema_file) as file:
            schema_data = json.load(file)

        return schema_data
    
    @classmethod
    def from_yml(
            cls,
            schema_file: str,
    )-> dict:
        """
        The attributes are extracted from the YAML file.
        The schema name and description are extracted from the YAML file if it contains them, overwriting the input parameters.

        The format of the yaml has to be in the form:
        schema:
          name:
          description:
          fields:
            - name: attribute_name
              description: description
        ...
        """

        # Load the schema from the YAML file
        with open(schema_file) as file:
            schema_data = yaml.safe_load(file)

        schema_data = schema_data["schema"]
        return {
            "name": schema_data.get("name", ""),
            "description": schema_data.get("description", ""),
            "fields": schema_data.get("fields", []),
            "type": schema_data.get("type", "")
        }


import logging

from palimpzest.core.data.dataclasses import ExecutionStats
from palimpzest.core.elements.records import DataRecordCollection
from palimpzest.query.processor.query_processor import QueryProcessor

logger = logging.getLogger(__name__)

class NoSentinelQueryProcessor(QueryProcessor):
    """
    Query processor that uses naive cost estimates to select the best plan.
    """

    # TODO: Consider to support dry_run.
    def execute(self) -> DataRecordCollection:
        logger.info("Executing NoSentinelQueryProcessor")

        # create execution stats
        execution_stats = ExecutionStats(execution_id=self.execution_id())
        execution_stats.start()

        # execute plan(s) according to the optimization strategy
        records, plan_stats = self._execute_best_plan(self.dataset, self.optimizer)

        # update the execution stats to account for the work to execute the final plan
        execution_stats.add_plan_stats(plan_stats)
        execution_stats.finish()

        # construct and return the DataRecordCollection
        result = DataRecordCollection(records, execution_stats=execution_stats)
        logger.info("Done executing NoSentinelQueryProcessor")

        return result



import time
from abc import ABC, abstractmethod
from dataclasses import dataclass

from chromadb.api.models.Collection import Collection
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    SpinnerColumn,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.progress import Progress as RichProgress
from rich.table import Table

from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import LLMFilter
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.retrieve import RetrieveOp
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan


@dataclass
class ProgressStats:
    """Statistics tracked for progress reporting"""
    start_time: float = 0.0
    total_cost: float = 0.0
    success_count: int = 0
    failure_count: int = 0
    current_operation: str = ""
    memory_usage_mb: float = 0.0
    recent_text: str = ""

def get_memory_usage() -> float:
    """Get current memory usage in MB"""
    try:
        import psutil
        process = psutil.Process()
        return process.memory_info().rss / 1024 / 1024
    except Exception:
        return 0.0

# NOTE: right now we only need to support single plan execution; in a multi-plan setting, we will
#       need to modify the semantics of the progress manager to support multiple plans
class ProgressManager(ABC):
    """Abstract base class for progress managers for plan execution"""

    def __init__(self, plan: PhysicalPlan | SentinelPlan, num_samples: int | None = None):
        """
        Initialize the progress manager for the given plan. This function takes in a plan,
        the number of samples to process (if specified).

        If `num_samples` is None, then the entire DataReader will be scanned.

        For each operator which is not an `AggregateOp` or `LimitScanOp`, we set its task `total`
        to the number of inputs to be processed by the plan. As intermediate operators process
        their inputs, the ProgressManager will update the `total` for their downstream operators.
        """
        # initialize progress object
        self.progress = RichProgress(
            SpinnerColumn(),
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            #TextColumn("[green]Success: {task.fields[success]}"),
            #TextColumn("[red]Failed: {task.fields[failed]}"),
            #TextColumn("[cyan]Mem: {task.fields[memory]:.1f}MB"),
            TextColumn("[green]Cost: ${task.fields[cost]:.4f}"),
            TextColumn("\n[white]{task.fields[recent]}"),  # Recent text on new line
            refresh_per_second=10,
            expand=True,   # Use full width
        )

        # initialize mapping from full_op_id --> ProgressStats
        self.full_op_id_to_stats: dict[str, ProgressStats] = {}

        # initialize mapping from full_op_id --> task
        self.full_op_id_to_task = {}

        # initialize start time
        self.start_time = None

        # create mapping from full_op_id --> next_op
        self.full_op_id_to_next_op: dict[str, PhysicalOperator] = {}
        for op_idx, op in enumerate(plan.operators):
            full_op_id = op.get_full_op_id()
            next_op = plan.operators[op_idx + 1] if op_idx + 1 < len(plan.operators) else None
            self.full_op_id_to_next_op[full_op_id] = next_op

        # compute the total number of inputs to be processed by the plan
        datareader_len = len(plan.operators[0].datareader)
        total = datareader_len if num_samples is None else min(num_samples, datareader_len)

        # add a task to the progress manager for each operator in the plan
        for op in plan.operators:
            # get the op id and a short string representation of the op; (str(op) is too long)
            op_str = f"{op.op_name()} ({op.get_op_id()})"

            # update the `total` if we encounter an AggregateOp or LimitScanOp
            if isinstance(op, AggregateOp):
                total = 1
            elif isinstance(op, LimitScanOp):
                total = op.limit

            self.add_task(op.get_full_op_id(), op_str, total)

    def get_task_total(self, full_op_id: str) -> int:
        """Return the current total value for the given task."""
        task = self.full_op_id_to_task[full_op_id]
        return self.progress._tasks[task].total

    def get_task_description(self, full_op_id: str) -> str:
        """Return the current description for the given task."""
        task = self.full_op_id_to_task[full_op_id]
        return self.progress._tasks[task].description

    @abstractmethod
    def add_task(self, full_op_id: str, op_str: str, total: int):
        """Initialize progress tracking for operator execution with total items"""
        pass

    @abstractmethod
    def start(self):
        """Start the progress bar(s)"""
        pass

    @abstractmethod
    def incr(self, full_op_id: str, num_outputs: int = 1, display_text: str | None = None, **kwargs):
        """
        Advance the progress bar for the given operator by one. Modify the downstream operators'
        progress bar `total` to reflect the number of outputs produced by this operator.

        NOTE: The semantics of this function are that every time it is executed we advance the
        progress bar by 1. This is because the progress bar represents what fraction of the inputs
        have been processed by the operator. `num_outputs` specifies how many outputs were generated
        by the operator when processing the input for which `incr()` was called. E.g. a filter which
        filters an input record will advance its progress bar by 1, but the next operator will now
        have 1 fewer inputs to process. Alternatively, a convert which generates 3 `num_outputs` will
        increase the inputs for the next operator by `delta = num_outputs - 1 = 2`.
        """
        pass

    @abstractmethod
    def finish(self):
        """Clean up and finalize progress tracking"""
        pass


class MockProgressManager(ProgressManager):
    """Mock progress manager for testing purposes"""

    def __init__(self, plan: PhysicalPlan | SentinelPlan, num_samples: int | None = None):
        pass

    def add_task(self, full_op_id: str, op_str: str, total: int):
        pass

    def start(self):
        pass

    def incr(self, full_op_id: str, num_outputs: int = 1, display_text: str | None = None, **kwargs):
        pass

    def finish(self):
        pass

class PZProgressManager(ProgressManager):
    """Progress manager for command line interface using rich"""
    
    def __init__(self, plan: PhysicalPlan, num_samples: int | None = None):
        super().__init__(plan, num_samples)
        self.console = Console()

    def add_task(self, full_op_id: str, op_str: str, total: int):
        """Add a new task to the progress bar"""
        task = self.progress.add_task(
            f"[blue]{op_str}", 
            total=total,
            cost=0.0,
            success=0,
            failed=0,
            memory=0.0,
            recent="",
        )

        # store the mapping of operator ID to task ID
        self.full_op_id_to_task[full_op_id] = task

        # initialize the stats for this operation
        self.full_op_id_to_stats[full_op_id] = ProgressStats(start_time=time.time())

    def start(self):
        # print a newline before starting to separate from previous output
        print()

        # set start time
        self.start_time = time.time()

        # start progress bar
        self.progress.start()

    def incr(self, full_op_id: str, num_outputs: int = 1, display_text: str | None = None, **kwargs):
        # get the task for the given operation
        task = self.full_op_id_to_task.get(full_op_id)

        # update statistics with any additional keyword arguments
        if kwargs != {}:
            self.update_stats(full_op_id, **kwargs)

        # update progress bar and recent text in one update
        if display_text is not None:
            self.full_op_id_to_stats[full_op_id].recent_text = display_text

        # if num_outputs is not 1, update the downstream operators' progress bar total for any
        # operator which is not an AggregateOp or LimitScanOp
        delta = num_outputs - 1
        if delta != 0:
            next_op = self.full_op_id_to_next_op[full_op_id]
            while next_op is not None:
                if not isinstance(next_op, (AggregateOp, LimitScanOp)):
                    next_full_op_id = next_op.get_full_op_id()
                    next_task = self.full_op_id_to_task[next_full_op_id]
                    self.progress.update(next_task, total=self.get_task_total(next_full_op_id) + delta)

                next_op = self.full_op_id_to_next_op[next_full_op_id]

        # advance the progress bar for this task
        self.progress.update(
            task,
            advance=1,
            description=f"[bold blue]{self.get_task_description(full_op_id)}",
            cost=self.full_op_id_to_stats[full_op_id].total_cost,
            success=self.full_op_id_to_stats[full_op_id].success_count,
            failed=self.full_op_id_to_stats[full_op_id].failure_count,
            memory=get_memory_usage(),
            recent=f"{self.full_op_id_to_stats[full_op_id].recent_text}" if display_text is not None else "",
            refresh=True,
        )

    def finish(self):
        self.progress.stop()

        # compute total cost, success, and failure
        total_cost = sum(stats.total_cost for stats in self.full_op_id_to_stats.values())
        # success_count = sum(stats.success_count for stats in self.full_op_id_to_stats.values())
        # failure_count = sum(stats.failure_count for stats in self.full_op_id_to_stats.values())

        # Print final stats on new lines after progress display
        print(f"Total time: {time.time() - self.start_time:.2f}s")
        print(f"Total cost: ${total_cost:.4f}")
        # print(f"Success rate: {success_count}/{success_count + failure_count}")

    def update_stats(self, full_op_id: str, **kwargs):
        """Update progress statistics"""
        for key, value in kwargs.items():
            if hasattr(self.full_op_id_to_stats[full_op_id], key):
                if key != "total_cost":
                    setattr(self.full_op_id_to_stats[full_op_id], key, value)
                else:
                    self.full_op_id_to_stats[full_op_id].total_cost += value
        self.full_op_id_to_stats[full_op_id].memory_usage_mb = get_memory_usage()

class PZSentinelProgressManager(ProgressManager):
    def __init__(self, plan: SentinelPlan, sample_budget: int):
        # overall progress bar
        self.overall_progress = RichProgress(
            SpinnerColumn(),
            TextColumn("{task.description}"),  # TODO: fixed string?
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            TextColumn("[green]Cost: ${task.fields[cost]:.4f}"),
            TextColumn("\n[white]{task.fields[recent]}"),  # Recent text on new line
            refresh_per_second=10,
            expand=True,   # Use full width
        )
        self.overall_task_id = self.overall_progress.add_task("", total=sample_budget, cost=0.0, recent="")

        # logical operator progress bars
        self.op_progress = RichProgress(
            SpinnerColumn(),
            "{task.description}",
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TextColumn("[green]Cost: ${task.fields[cost]:.4f}"),
            TextColumn("\n[white]{task.fields[recent]}"),  # Recent text on new line
            refresh_per_second=10,
            expand=True,   # Use full width
        )

        # organize progress bars into nice display
        self.progress_table = Table.grid()
        self.progress_table.add_row(
            Panel.fit(self.op_progress, title="[b]Sample Allocation", border_style="red", padding=(1, 2)),
        )
        self.progress_table.add_row(
            Panel.fit(
                self.overall_progress, title="Optimization Progress", border_style="green", padding=(2, 2)
            )
        )
        self.live_display = Live(self.progress_table, refresh_per_second=10)

        # initialize mapping from logical_op_id --> ProgressStats
        self.logical_op_id_to_stats: dict[str, ProgressStats] = {}

        # initialize mapping from logical_op_id --> task
        self.logical_op_id_to_task = {}

        # initialize start time
        self.start_time = None

        # add a task to the progress manager for each operator in the plan
        for logical_op_id, op_set in plan:
            physical_op = op_set[0]
            is_llm_convert = isinstance(physical_op, LLMConvert)
            is_llm_filter = isinstance(physical_op, LLMFilter)
            op_name = "LLMConvert" if is_llm_convert else "LLMFilter" if is_llm_filter else physical_op.op_name()
            op_str = f"{op_name} ({logical_op_id})"
            total = sample_budget if self._is_llm_op(op_set[0]) else 0
            self.add_task(logical_op_id, op_str, total)

        self.console = Console()

    def _is_llm_op(self, physical_op: PhysicalOperator) -> bool:
        is_llm_convert = isinstance(physical_op, LLMConvert)
        is_llm_filter = isinstance(physical_op, LLMFilter)
        is_llm_retrieve = isinstance(physical_op, RetrieveOp) and isinstance(physical_op.index, Collection)
        return is_llm_convert or is_llm_filter or is_llm_retrieve

    def get_task_description(self, logical_op_id: str) -> str:
        """Return the current description for the given task."""
        task = self.logical_op_id_to_task[logical_op_id]
        return self.op_progress._tasks[task].description

    def add_task(self, logical_op_id: str, op_str: str, total: int):
        """Add a new task to the op progress bars"""
        task = self.op_progress.add_task(
            f"[blue]{op_str}", 
            total=total,
            cost=0.0,
            success=0,
            failed=0,
            memory=0.0,
            recent="",
        )

        # store the mapping of operator ID to task ID
        self.logical_op_id_to_task[logical_op_id] = task

        # initialize the stats for this operation
        self.logical_op_id_to_stats[logical_op_id] = ProgressStats(start_time=time.time())

    def start(self):
        # print a newline before starting to separate from previous output
        print()

        # set start time
        self.start_time = time.time()

        # start progress bars
        self.live_display.start()

    def incr(self, logical_op_id: str, num_samples: int, display_text: str | None = None, **kwargs):
        # TODO: (above) organize progress bars into a Live / Table / Panel or something
        # get the task for the given operation
        task = self.logical_op_id_to_task.get(logical_op_id)

        # update statistics with any additional keyword arguments
        if kwargs != {}:
            self.update_stats(logical_op_id, **kwargs)

        # update progress bar and recent text in one update
        if display_text is not None:
            self.logical_op_id_to_stats[logical_op_id].recent_text = display_text

        # advance the op progress bar for this logical_op_id
        self.op_progress.update(
            task,
            advance=num_samples,
            description=f"[bold blue]{self.get_task_description(logical_op_id)}",
            cost=self.logical_op_id_to_stats[logical_op_id].total_cost,
            success=self.logical_op_id_to_stats[logical_op_id].success_count,
            failed=self.logical_op_id_to_stats[logical_op_id].failure_count,
            memory=get_memory_usage(),
            recent=f"{self.logical_op_id_to_stats[logical_op_id].recent_text}" if display_text is not None else "",
            refresh=True,
        )

        # advance the overall progress bar
        self.overall_progress.update(
            self.overall_task_id,
            advance=num_samples,
            cost=sum(stats.total_cost for _, stats in self.logical_op_id_to_stats.items()),
            refresh=True,
        )

        # force the live display to refresh
        self.live_display.refresh()

    def finish(self):
        self.live_display.stop()

        # compute total cost, success, and failure
        total_cost = sum(stats.total_cost for stats in self.logical_op_id_to_stats.values())
        # success_count = sum(stats.success_count for stats in self.logical_op_id_to_stats.values())
        # failure_count = sum(stats.failure_count for stats in self.logical_op_id_to_stats.values())

        # Print final stats on new lines after progress display
        print(f"Total opt. time: {time.time() - self.start_time:.2f}s")
        print(f"Total opt. cost: ${total_cost:.4f}")
        # print(f"Success rate: {success_count}/{success_count + failure_count}")

    def update_stats(self, logical_op_id: str, **kwargs):
        """Update progress statistics"""
        for key, value in kwargs.items():
            if hasattr(self.logical_op_id_to_stats[logical_op_id], key):
                if key != "total_cost":
                    setattr(self.logical_op_id_to_stats[logical_op_id], key, value)
                else:
                    self.logical_op_id_to_stats[logical_op_id].total_cost += value
        self.logical_op_id_to_stats[logical_op_id].memory_usage_mb = get_memory_usage()

def create_progress_manager(
    plan: PhysicalPlan | SentinelPlan,
    num_samples: int | None = None,
    sample_budget: int | None = None,
    progress: bool = True,
) -> ProgressManager:
    """Factory function to create appropriate progress manager based on environment"""
    if not progress:
        return MockProgressManager(plan, num_samples)

    if isinstance(plan, SentinelPlan):
        assert sample_budget is not None, "Sample budget must be specified for SentinelPlan progress manager"
        return PZSentinelProgressManager(plan, sample_budget)

    return PZProgressManager(plan, num_samples)



"""This file contains prompts for SplitConvert operations on text inputs."""

### BASE PROMPTS ###
COT_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a paragraph or two which describes what you believe is the correct value for each output field.
Be sure to cite information from the context as evidence of why your answers are correct. Do not hallucinate evidence.

You will be provided with a description of each input field and each output field.

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

OUTPUT FIELDS:
{example_output_fields}

CONTEXT:
{example_context}

Let's think step-by-step in order to answer the question.

ANSWER: {example_answer}
---
"""

COT_SPLIT_PROPOSER_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a paragraph or two which describes what you believe is the correct value for each output field.
Be sure to cite information from the context as evidence of why your answers are correct. Do not hallucinate evidence.

You will be provided with a description of each input field and each output field.
---
INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

CONTEXT:
{context}

Let's think step-by-step in order to answer the question.

ANSWER: """


### TEMPLATE INPUTS ###
SPLIT_PROPOSER_JOB_INSTRUCTION = "produce an answer to a question"
SPLIT_PROPOSER_EXAMPLE_INPUT_FIELDS = """- text: a text passage describing scientists"""
SPLIT_PROPOSER_EXAMPLE_OUTPUT_FIELDS = """- name: the list of names for each scientist mentioned in the text
- field_of_study: a list with the field of study for each scientist"""
SPLIT_PROPOSER_EXAMPLE_CONTEXT = """{{
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, born December 10, 1815 was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation."
}}"""
SPLIT_PROPOSER_EXAMPLE_ANSWER = """the text passage mentions the scientists "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace" and "Charles Babbage", both of whom were mathematicians. Therefore, the name output should be ["Augusta Ada King", "Charles Babbage"] and the field_of_study output should be ["Mathematician", "Mathematician"]."""



import contextlib
import io

import IPython.terminal.embed as embed


class Sandbox:
    def __init__(self, codes: list | None = None, reset: bool = False):
        if codes is None:
            codes = list()
        super().__init__()
        self.shell = None
        self.codes = codes
        if reset:
            self.reset()

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exit()

    def reset(self):
        self.exit()
        self.shell = embed.InteractiveShellEmbed(quiet=True)
        for code in self.codes:
            response = self.shell.run_cell(code)
            if not response.success:
                return {"status": False, "response": None, "msg": response["msg"]}
        return {"status": True, "response": self.shell.user_ns["_"], "msg": None}

    def execute(self, code, reset=False):
        try:
            if reset:
                response = self.reset()
                if not response["status"]:
                    return {"status": False, "response": None, "msg": str(response.error_in_exec)}
            with io.StringIO() as buf, contextlib.redirect_stdout(buf), contextlib.redirect_stderr(buf):
                response = self.shell.run_cell(code)
            if response.success:
                return {"status": True, "response": self.shell.user_ns["_"], "msg": None}
            else:
                return {"status": False, "response": None, "msg": str(response.error_in_exec)}
        except Exception as e:
            return {"status": False, "response": None, "msg": str(type(e)) + " " + str(e)}

    def add(self, code, index=-1):
        self.codes.insert(index, code)

    def pop(self, index=-1):
        return self.codes.pop(index)

    def get(self, key="_"):
        return self.shell.user_ns[key] if self.shell else None

    def export(self, path):
        with open(path + ".py", "w") as f:
            for code in self.codes:
                f.write("# %%\n")
                f.write(code)
                f.write("\n\n")
            f.write("# %%\n")

    def exit(self):
        if self.shell:
            self.shell.run_cell("quit()")
        self.shell = None

    @classmethod
    def get_globals(cls):
        with cls(codes=[]) as env:
            response = env.execute("globals()")
        if response["status"]:
            return response["response"]
        else:
            return {}


class API:
    def __init__(self, **kwargs):
        super().__init__()
        self.config = kwargs
        assert "name" in self.config, "Name is required!"
        assert "inputs" in self.config, "Inputs are required!"
        assert "outputs" in self.config, "Outputs are required!"
        assert len(self.config["outputs"]) == 1, "Currently only single output is supported!"

        if "env" not in self.config:
            self.env = Sandbox(codes=list())
            # self.env = Sandbox(codes=["from seed import *"])
        else:
            self.env = Sandbox(codes=self.config["env"])

    @classmethod
    def from_task_descriptor(cls, td, field_name, input_fields=None):
        name, inputs, outputs = "extraction", list(), list()
        if input_fields is None:
            input_fields = td.input_schema.field_names()
        for input_field_name in input_fields:
            inputs.append({"name": input_field_name, "desc": getattr(td.input_schema, input_field_name).desc})
        outputs = [{"name": field_name, "desc": getattr(td.output_schema, field_name).desc}]
        return cls(name=name, inputs=inputs, outputs=outputs)

    @classmethod
    def from_input_output_schemas(cls, input_schema, output_schema, field_name, input_fields=None):
        name, inputs, outputs = "extraction", list(), list()
        if input_fields is None:
            input_fields = input_schema.field_names()
        for input_field_name in input_fields:
            inputs.append({"name": input_field_name, "desc": getattr(input_schema, input_field_name).desc})
        outputs = [{"name": field_name, "desc": getattr(output_schema, field_name).desc}]
        return cls(name=name, inputs=inputs, outputs=outputs)

    @property
    def name(self):
        return self.config["name"]

    @property
    def inputs(self):
        return [i["name"] for i in self.config["inputs"]]

    @property
    def input_descs(self):
        return [i["desc"] for i in self.config["inputs"]]

    @property
    def output(self):
        # Currently only single output is supported!
        return self.config["outputs"][0]["name"]

    @property
    def output_desc(self):
        # Currently only single output is supported!
        return self.config["outputs"][0]["desc"]

    @property
    def doc(self):
        return self.config.get("doc", "")

    @property
    def args(self):
        # attr1, attr2, attr3, ...
        return ", ".join(i["name"] for i in self.config["inputs"])

    @property
    def asgs(self):
        # attr1=attr1, attr2=attr2, attr3=attr3, ...
        return ", ".join(i["name"] + "=" + i["name"] for i in self.config["inputs"])

    def inps(self, inputs=None):
        # attr1=..., attr2=..., attr3=..., ...
        if inputs is None:
            inputs = dict()
        return ", ".join(
            i["name"] + "=" + repr(inputs[i["name"]]) for i in self.config["inputs"] if i["name"] in inputs
        )

    def kwargs_call(self):
        # name(**kwargs)
        return self.config["name"] + "(**kwargs)"

    def args_call(self, with_kwargs=False):
        # name(attr1, attr2, attr3, ..., **kwargs)
        return self.config["name"] + "(" + self.args + (", **kwargs)" if with_kwargs else ")")

    def asgs_call(self, with_kwargs=False):
        # name(attr1=attr1, attr2=attr2, attr3=attr3, ..., **kwargs)
        return self.config["name"] + "(" + self.asgs + (", **kwargs)" if with_kwargs else ")")

    def api_def(self, with_kwargs=False):
        # def name(attr1, attr2, attr3, ..., **kwargs):
        return "def " + self.args_call(with_kwargs=with_kwargs) + ":\n"

    def api_call(self, inputs, with_kwargs=False):
        # name(attr1=..., attr2=..., attr3=..., ..., **kwargs)
        return self.config["name"] + "(" + self.inps(inputs) + (", **kwargs)" if with_kwargs else ")")

    def api_execute(self, code, inputs):
        self.env.add(code)
        result = self.env.execute(self.api_call(inputs), reset=True)
        self.env.pop()
        return result



from __future__ import annotations

import logging

# NOTE: the answer.mode() call(s) inside of _est_quality() throw a UserWarning when there are multiple
#       answers to a convert with the same mode. This is because pandas tries to sort the answers
#       before returning them, but since answer is a column of dicts the '<' operator fails on dicts.
#       For now, we can simply ignore the warning b/c we pick an answer at random anyways if there are
#       multiple w/the same count, but in the future we may want to cast the 'dict' --> 'str' or compute
#       the mode on a per-field basis.
import warnings
from typing import Any

import pandas as pd

from palimpzest.constants import MODEL_CARDS, NAIVE_BYTES_PER_RECORD, GPT_4o_MODEL_CARD, Model
from palimpzest.core.data.dataclasses import OperatorCostEstimates, PlanCost, RecordOpStats, SentinelPlanStats
from palimpzest.query.operators.aggregate import ApplyGroupByOp, AverageAggregateOp, CountAggregateOp
from palimpzest.query.operators.code_synthesis_convert import CodeSynthesisConvert
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.filter import LLMFilter, NonLLMFilter
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.rag_convert import RAGConvert
from palimpzest.query.operators.scan import CacheScanDataOp, MarshalAndScanDataOp, ScanPhysicalOp
from palimpzest.utils.model_helpers import get_champion_model_name, get_models

warnings.simplefilter(action='ignore', category=UserWarning)

logger = logging.getLogger(__name__)

class BaseCostModel:
    """
    This base class contains the interface/abstraction that every CostModel must implement
    in order to work with the Optimizer. In brief, the Optimizer expects the CostModel to
    make a prediction about the runtime, cost, and quality of a physical operator.
    """
    def __init__(self):
        """
        CostModel constructor; the arguments for individual CostModels may vary depending
        on the assumptions they make about the prevalance of historical execution data
        and online vs. batch execution settings.
        """
        pass

    def get_costed_full_op_ids(self) -> set[str]:
        """
        Return the set of full op ids which the cost model has cost estimates for.
        """
        raise NotImplementedError("Calling get_costed_full_op_ids from abstract method")

    def __call__(self, operator: PhysicalOperator) -> PlanCost:
        """
        The interface exposed by the CostModel to the Optimizer. Subclasses may require
        additional arguments in order to make their predictions.
        """
        raise NotImplementedError("Calling __call__ from abstract method")


class SampleBasedCostModel:
    """
    """
    def __init__(
        self,
        sentinel_plan_stats: SentinelPlanStats,
        verbose: bool = False,
        exp_name: str | None = None,
    ):
        # store verbose argument
        self.verbose = verbose

        # store experiment name if one is provided
        self.exp_name = exp_name

        # construct cost, time, quality, and selectivity matrices for each operator set;
        self.operator_to_stats = self.compute_operator_stats(sentinel_plan_stats)
        self.costed_full_op_ids = set([
            full_op_id
            for _, full_op_id_to_stats in self.operator_to_stats.items()
            for full_op_id in full_op_id_to_stats
        ])

        logger.info(f"Initialized SampleBasedCostModel with verbose={self.verbose}")
        logger.debug(f"Initialized SampleBasedCostModel with params: {self.__dict__}")

    def get_costed_full_op_ids(self):
        return self.costed_full_op_ids

    def compute_operator_stats(self, sentinel_plan_stats: SentinelPlanStats) -> dict:
        logger.debug("Computing operator statistics")
        # flatten the nested dictionary of execution data and pull out fields relevant to cost estimation
        execution_record_op_stats = []
        for logical_op_id, full_op_id_to_op_stats in sentinel_plan_stats.operator_stats.items():
            logger.debug(f"Computing operator statistics for logical_op_id: {logical_op_id}")
            # flatten the execution data into a list of RecordOpStats
            op_set_execution_data = [
                record_op_stats
                for _, op_stats in full_op_id_to_op_stats.items()
                for record_op_stats in op_stats.record_op_stats_lst
            ]

            # add entries from execution data into matrices
            for record_op_stats in op_set_execution_data:
                record_op_stats_dict = {
                    "logical_op_id": logical_op_id,
                    "full_op_id": record_op_stats.full_op_id,
                    "record_id": record_op_stats.record_id,
                    "record_parent_id": record_op_stats.record_parent_id,
                    "cost_per_record": record_op_stats.cost_per_record,
                    "time_per_record": record_op_stats.time_per_record,
                    "quality": record_op_stats.quality,
                    "passed_operator": record_op_stats.passed_operator,
                    "source_idx": record_op_stats.record_source_idx,  # TODO: remove
                    "op_details": record_op_stats.op_details,         # TODO: remove
                    "answer": record_op_stats.answer,                 # TODO: remove
                }
                execution_record_op_stats.append(record_op_stats_dict)

        # convert flattened execution data into dataframe
        operator_stats_df = pd.DataFrame(execution_record_op_stats)

        # for each full_op_id, compute its average cost_per_record, time_per_record, selectivity, and quality
        operator_to_stats = {}
        for logical_op_id, logical_op_df in operator_stats_df.groupby("logical_op_id"):
            logger.debug(f"Computing operator statistics for logical_op_id: {logical_op_id}")
            operator_to_stats[logical_op_id] = {}

            for full_op_id, physical_op_df in logical_op_df.groupby("full_op_id"):
                # compute the number of input records processed by this operator; use source_idx for scan operator(s)
                num_source_records = (
                    len(physical_op_df.record_parent_id.unique())
                    if not physical_op_df.record_parent_id.isna().all()
                    else len(physical_op_df.source_idx.unique())
                )

                # compute selectivity
                selectivity = physical_op_df.passed_operator.sum() / num_source_records

                operator_to_stats[logical_op_id][full_op_id] = {
                    "cost": physical_op_df.cost_per_record.mean(),
                    "time": physical_op_df.time_per_record.mean(),
                    "quality": physical_op_df.quality.mean(),
                    "selectivity": selectivity,
                }

        # if this is an experiment, log the dataframe and operator_to_stats dictionary
        if self.exp_name is not None:
            operator_stats_df.to_csv(f"opt-profiling-data/{self.exp_name}-operator-stats.csv", index=False)

        logger.debug(f"Done computing operator statistics for {len(operator_to_stats)} operators!")
        return operator_to_stats


    def __call__(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # NOTE: some physical operators may not have any sample execution data in this cost model;
        #       these physical operators are filtered out of the Optimizer, thus we can assume that
        #       we will have execution data for each operator passed into __call__; nevertheless, we
        #       still perform a sanity check
        # look up physical and logical op ids associated with this physical operator
        full_op_id = operator.get_full_op_id()
        logical_op_id = operator.logical_op_id
        physical_op_to_stats = self.operator_to_stats.get(logical_op_id)
        assert physical_op_to_stats is not None, f"No execution data for logical operator: {str(operator)}"
        assert physical_op_to_stats.get(full_op_id) is not None, f"No execution data for physical operator: {str(operator)}"
        logger.debug(f"Calling __call__ for {str(operator)}")

        # look up stats for this operation
        est_cost_per_record = self.operator_to_stats[logical_op_id][full_op_id]["cost"]
        est_time_per_record = self.operator_to_stats[logical_op_id][full_op_id]["time"]
        est_quality = self.operator_to_stats[logical_op_id][full_op_id]["quality"]
        est_selectivity = self.operator_to_stats[logical_op_id][full_op_id]["selectivity"]

        # create source_op_estimates for scan operators if they are not provided
        if isinstance(operator, ScanPhysicalOp):
            # get handle to scan operator and pre-compute its size (number of records)
            datareader_len = len(operator.datareader)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datareader_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

        # generate new set of OperatorCostEstimates
        op_estimates = OperatorCostEstimates(
            cardinality=est_selectivity * source_op_estimates.cardinality,
            time_per_record=est_time_per_record,
            cost_per_record=est_cost_per_record,
            quality=est_quality,
        )

        # compute estimates for this operator
        op_time = op_estimates.time_per_record * source_op_estimates.cardinality
        op_cost = op_estimates.cost_per_record * source_op_estimates.cardinality
        op_quality = op_estimates.quality

        # construct and return op estimates
        plan_cost = PlanCost(cost=op_cost, time=op_time, quality=op_quality, op_estimates=op_estimates)
        logger.debug(f"Done calling __call__ for {str(operator)}")
        logger.debug(f"Plan cost: {plan_cost}")
        return plan_cost


class CostModel(BaseCostModel):
    """
    This class takes in a list of RecordOpStats and performs cost estimation on a given operator
    by taking the average of any sample execution that the CostModel has for that operator. If no
    such data exists, it returns a naive estimate.
    """
    def __init__(
        self,
        sample_execution_data: list[RecordOpStats] | None = None,
        available_models: list[Model] | None = None,
    ) -> None:
        if sample_execution_data is None:
            sample_execution_data = []
        if available_models is None:
            available_models = []

        # construct full dataset of samples
        self.sample_execution_data_df = (
            pd.DataFrame(sample_execution_data)
            if len(sample_execution_data) > 0
            else None
        )
        # df contains a column called record_state, that sometimes contain a dict
        # we want to extract the keys from the dict and create a new column for each key

        # set available models
        self.available_models = available_models

        # compute per-operator estimates
        self.operator_estimates = self._compute_operator_estimates()

        # compute set of costed full op ids from operator_to_stats
        self.costed_full_op_ids = None if self.operator_estimates is None else set(self.operator_estimates.keys())
        logger.info("Initialized CostModel.")
        logger.debug(f"Initialized CostModel with params: {self.__dict__}")

    def get_costed_full_op_ids(self):
        return self.costed_full_op_ids

    def _compute_mean(self, df: pd.DataFrame, col: str, model_name: str | None = None) -> float:
        """
        Compute the mean for the given column and dataframe. If the model_name is provided, filter
        for the subset of rows belonging to the model.
        """
        # use model-specific estimate if possible
        if model_name is not None:
            model_df = df[df.model_name == model_name]
            if not model_df.empty:
                return model_df[col].mean()

        # compute aggregate mean across all models
        return df[col].mean()

    def _est_time_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[float, float, float]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the time per record.
        """
        return self._compute_mean(df=op_df, col="time_per_record", model_name=model_name)

    def _est_cost_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[float, float, float]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the cost per record.
        """
        return self._compute_mean(df=op_df, col="cost_per_record", model_name=model_name)

    def _est_tokens_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[float, float]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the total input tokens and total output tokens.
        """
        total_input_tokens = self._compute_mean(df=op_df, col="total_input_tokens", model_name=model_name)
        total_output_tokens = self._compute_mean(df=op_df, col="total_output_tokens", model_name=model_name)

        return total_input_tokens, total_output_tokens

    def _est_cardinality(self, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for a specific operation, compute the number of
        rows output by the operation.

        NOTE: right now, this should only be used by the ApplyGroupByOp as a way to gauge the
        number of output groups. Using this to estimate, e.g. the cardinality of a filter,
        convert, or base scan will lead to wildly inaccurate results because the absolute value
        of these cardinalities will simply be a reflection of the sample size.

        For those operations, we use the `_est_selectivity` function to estimate the operator's
        selectivity, which we can apply to an est. of the operator's input cardinality.
        """
        return op_df.shape[0] / len(op_df.plan_id.unique())

    def _est_selectivity(self, df: pd.DataFrame, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for the plan and a specific operation, compute
        the ratio of records between this operator and its source operator.
        """
        # compute whether or not this operation is a filter
        is_filter_op = "filter" in str(op_df.op_name.iloc[0]).lower()

        # use model-specific estimate if possible
        if model_name is not None:
            model_op_df = op_df[op_df.model_name == model_name]
            if not model_op_df.empty:
                num_input_records = model_op_df.shape[0]

                # get subset of records that were the source to this operator
                num_output_records = None
                if is_filter_op:
                    num_output_records = model_op_df.passed_operator.sum()
                else:
                    full_op_ids = model_op_df.full_op_id.unique().tolist()
                    plan_ids = model_op_df.plan_id.unique().tolist()
                    num_output_records = df[df.source_full_op_id.isin(full_op_ids) & df.plan_id.isin(plan_ids)].shape[0]

                # estimate the selectivity / fan-out
                return num_output_records / num_input_records

        # otherwise average selectivity across all ops
        num_input_records = op_df.shape[0]

        # get subset of records that were the source to this operator
        num_output_records = None
        if is_filter_op:
            num_output_records = op_df.passed_operator.sum()
        else:
            full_op_ids = op_df.full_op_id.unique().tolist()
            num_output_records = df[df.source_full_op_id.isin(full_op_ids)].shape[0]

        # estimate the selectivity / fan-out
        return num_output_records / num_input_records

    def _compute_quality(self, row):
        # compute accuracy for filter
        if "filter" in row["op_name"].lower():
            row["correct"] = int(row["answer"] == row["accepted_answer"])
            row["num_answers"] = 1
            return row

        # otherwise, compute recall on a per-key basis
        try:
            # we'll measure recall on accepted_answer, as extraneous info is often not an issue
            answer = row["answer"]
            accepted_answer = row["accepted_answer"]
            correct = 0
            for key, value in accepted_answer.items():
                if key in answer and answer[key] == value:
                    correct += 1
            
            row["correct"] = correct
            row["num_answers"] = len(accepted_answer.keys())
            return row

        except Exception as e:
            print(f"WARNING: error decoding answer or accepted_answer: {str(e)}")
            row["correct"] = 0
            row["num_answers"] = 1
            return row

    def _est_quality(self, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for a specific operation, compute the an estimate
        of the quality of its outputs by using GPT-4 as a champion model.
        """
        # get unique set of records
        record_ids = op_df.record_id.unique()

        # get champion model name
        vision = ("image_operation" in op_df.columns and op_df.image_operation.any())
        champion_model_name = get_champion_model_name(self.available_models, vision)

        # compute champion's answer (per-record) across all models; fall-back to most common answer if champion is not present
        record_id_to_answer = {}
        for record_id in record_ids:
            record_df = op_df[op_df.record_id == record_id]
            champion_most_common_answer = record_df[
                record_df.model_name == champion_model_name
            ].answer.mode()
            all_models_most_common_answer = record_df.answer.mode()

            if not champion_most_common_answer.empty:
                record_id_to_answer[record_id] = champion_most_common_answer.iloc[0]
            elif not all_models_most_common_answer.empty:
                record_id_to_answer[record_id] = all_models_most_common_answer.iloc[0]
            else:
                record_id_to_answer[record_id] = None

        # compute accepted answers and clean all answers
        pd.options.mode.chained_assignment = None  # turn off copy warnings
        op_df.loc[:, "accepted_answer"] = op_df.record_id.apply(lambda id: record_id_to_answer[id])
        op_df = op_df.apply(lambda row: self._compute_quality(row), axis=1)

        # get subset of observations for model_name and estimate quality w/fraction of answers that match accepted answer
        model_df = (
            op_df[op_df.model_name == model_name]
            if model_name is not None
            else op_df[op_df.model_name.isna()]
        )

        # compute quality as the fraction of answers which are correct (recall on expected output)
        num_correct = model_df.correct.sum() if not model_df.empty else op_df.correct.sum()
        total_answers = model_df.num_answers.sum() if not model_df.empty else op_df.num_answers.sum()
        est_quality = num_correct / total_answers

        return est_quality

    def _compute_operator_estimates(self) -> dict[str, Any] | None:
        """
        Compute per-operator estimates of runtime, cost, and quality.
        """
        # if we don't have sample execution data, we cannot compute per-operator estimates
        if self.sample_execution_data_df is None:
            return None

        # get the set of operator ids for which we have sample data
        full_op_ids = self.sample_execution_data_df.full_op_id.unique()

        # compute estimates of runtime, cost, and quality (and intermediates like cardinality) for every operator
        operator_estimates = {}
        for full_op_id in full_op_ids:
            # filter for subset of sample execution data related to this operation
            op_df = self.sample_execution_data_df[
                self.sample_execution_data_df.full_op_id == full_op_id
            ]

            # skip computing an estimate if we didn't capture any sampling data for this operator
            # (this can happen if/when upstream filter operation(s) filter out all records) 
            if op_df.empty:
                continue

            # initialize estimates
            estimates = {}

            # get the op_name for this operation
            model_name = op_df.model_name.iloc[0] if op_df.model_name.iloc[0] is not None else None
            op_name = str(op_df.op_name.iloc[0])
            if model_name is not None:
                # compute estimates per-model, and add None which forces computation of avg. across all models
                model_names = [m.value for m in get_models(include_vision=True)] + [None]
                # model_names = op_df.model_name.unique().tolist()
                estimates = {model_name: None for model_name in model_names}
                for model_name in model_names:
                    time_per_record = self._est_time_per_record(op_df, model_name=model_name)
                    cost_per_record = self._est_cost_per_record(op_df, model_name=model_name)
                    input_tokens, output_tokens = self._est_tokens_per_record(op_df, model_name=model_name)
                    selectivity = self._est_selectivity(self.sample_execution_data_df, op_df, model_name=model_name)
                    quality = self._est_quality(op_df, model_name=model_name)

                    model_estimates = {
                        "time_per_record": time_per_record,
                        "cost_per_record": cost_per_record,
                        "total_input_tokens": input_tokens,
                        "total_output_tokens": output_tokens,
                        "selectivity": selectivity,
                        "quality": quality,
                    }
                    estimates[model_name] = model_estimates

            # TODO pre-compute lists of op_names in groups
            elif op_name in ["NonLLMFilter"]:
                time_per_record = self._est_time_per_record(op_df)
                selectivity = self._est_selectivity(self.sample_execution_data_df, op_df)
                estimates = {"time_per_record": time_per_record, "selectivity": selectivity}

            elif op_name in ["MarshalAndScanDataOp", "CacheScanDataOp", "LimitScanOp", "CountAggregateOp", "AverageAggregateOp"]:
                time_per_record = self._est_time_per_record(op_df)
                estimates = {"time_per_record": time_per_record}

            elif op_name in ["ApplyGroupByOp"]:
                time_per_record = self._est_time_per_record(op_df)
                cardinality = self._est_cardinality(op_df)
                estimates = {"time_per_record": time_per_record, "cardinality": cardinality}

            operator_estimates[full_op_id] = estimates

        return operator_estimates

    def __call__(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # get identifier for operation which is unique within sentinel plan but consistent across sentinels
        full_op_id = operator.get_full_op_id()
        logger.debug(f"Calling __call__ for {str(operator)} with full_op_id: {full_op_id}")

        # initialize estimates of operator metrics based on naive (but sometimes precise) logic
        if isinstance(operator, MarshalAndScanDataOp):
            # get handle to scan operator and pre-compute its size (number of records)
            datareader_len = len(operator.datareader)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datareader_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates, input_record_size_in_bytes=NAIVE_BYTES_PER_RECORD)

        elif isinstance(operator, CacheScanDataOp):
            datareader_len = len(operator.datareader)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datareader_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates, input_record_size_in_bytes=NAIVE_BYTES_PER_RECORD)

        else:
            op_estimates = operator.naive_cost_estimates(source_op_estimates)

        # if we have sample execution data, update naive estimates with more informed ones
        sample_op_estimates = self.operator_estimates
        if sample_op_estimates is not None and full_op_id in sample_op_estimates:
            if isinstance(operator, (MarshalAndScanDataOp, CacheScanDataOp)):
                op_estimates.time_per_record = sample_op_estimates[full_op_id]["time_per_record"]

            elif isinstance(operator, ApplyGroupByOp):
                # NOTE: in theory we should also treat this cardinality est. as a random variable, but in practice we will
                #       have K samples of the number of groups produced by the groupby operator, where K is the number of
                #       plans we generate sample data with. For now, we will simply use the estimate without bounds.
                #
                # NOTE: this cardinality is the only cardinality we estimate directly b/c we can observe how many groups are
                #       produced by the groupby in our sample and assume it may generalize to the full workload. To estimate
                #       actual cardinalities of operators we estimate their selectivities / fan-outs and multiply those by
                #       the input cardinality (where the initial input cardinality from the datareader is known).
                op_estimates.cardinality = sample_op_estimates[full_op_id]["cardinality"]
                op_estimates.time_per_record = sample_op_estimates[full_op_id]["time_per_record"]

            elif isinstance(operator, (CountAggregateOp, AverageAggregateOp)):  # noqa: SIM114
                op_estimates.time_per_record = sample_op_estimates[full_op_id]["time_per_record"]

            elif isinstance(operator, LimitScanOp):
                op_estimates.time_per_record = sample_op_estimates[full_op_id]["time_per_record"]

            elif isinstance(operator, NonLLMFilter):
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[full_op_id]["selectivity"]
                op_estimates.time_per_record = sample_op_estimates[full_op_id]["time_per_record"]

            elif isinstance(operator, LLMFilter):
                model_name = operator.model.value
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[full_op_id][model_name]["selectivity"]
                op_estimates.time_per_record = sample_op_estimates[full_op_id][model_name]["time_per_record"]
                op_estimates.cost_per_record = sample_op_estimates[full_op_id][model_name]["cost_per_record"]
                op_estimates.quality = sample_op_estimates[full_op_id][model_name]["quality"]

            elif isinstance(operator, LLMConvert):
                # NOTE: code synthesis does not have a model attribute
                model_name = operator.model.value if hasattr(operator, "model") else None
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[full_op_id][model_name]["selectivity"]
                op_estimates.time_per_record = sample_op_estimates[full_op_id][model_name]["time_per_record"]
                op_estimates.cost_per_record = sample_op_estimates[full_op_id][model_name]["cost_per_record"]
                op_estimates.quality = sample_op_estimates[full_op_id][model_name]["quality"]

                # NOTE: if code synth. fails, this will turn into ConventionalQuery calls to GPT-3.5,
                #       which would wildly mess up estimate of time and cost per-record
                # do code synthesis adjustment
                if isinstance(operator, CodeSynthesisConvert):
                    op_estimates.time_per_record = 1e-5
                    op_estimates.cost_per_record = 1e-4
                    op_estimates.quality = op_estimates.quality * (GPT_4o_MODEL_CARD["code"] / 100.0)

                # rag convert adjustment
                if isinstance(operator, RAGConvert):
                    total_input_tokens = operator.num_chunks_per_field * operator.chunk_size
                    total_output_tokens = sample_op_estimates[full_op_id][model_name]["total_output_tokens"]
                    op_estimates.cost_per_record = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens
                    )
                    op_estimates.quality = op_estimates.quality * operator.naive_quality_adjustment

            else:
                raise Exception("Unknown operator")

        # compute estimates for this operator
        op_time = op_estimates.time_per_record * source_op_estimates.cardinality
        op_cost = op_estimates.cost_per_record * source_op_estimates.cardinality
        op_quality = op_estimates.quality

        # create and return PlanCost object for this op's statistics
        op_plan_cost = PlanCost(
            cost=op_cost,
            time=op_time,
            quality=op_quality,
            op_estimates=op_estimates,
        )
        logger.debug(f"Done calling __call__ for {str(operator)} with full_op_id: {full_op_id}")
        logger.debug(f"Plan cost: {op_plan_cost}")

        return op_plan_cost



from __future__ import annotations

from abc import ABC, abstractmethod

from chromadb.api.models.Collection import Collection


def index_factory(index: Collection) -> PZIndex:
    """
    Factory function to create a PZ index based on the type of the provided index.

    Args:
        index (Collection): The index provided by the user.

    Returns:
        PZIndex: The PZ wrapped Index.
    """
    if isinstance(index, Collection):
        return ChromaIndex(index)
    else:
        raise TypeError(f"Unsupported index type: {type(index)}\nindex must be a `chromadb.api.models.Collection.Collection`")


class BaseIndex(ABC):

    def __init__(self, index: Collection):
        self.index = index

    def __str__(self):
        """
        Return a string representation of the index.
        """
        return f"{self.__class__.__name__}"

    @abstractmethod
    def search(self, query_embedding: list[float] | list[list[float]], results_per_query: int = 1) -> list | list[list]:
        """
        Query the index with a string or a list of strings.

        Args:
            query (str | list[str]): The query string or list of strings to search for.
            results_per_query (int): The number of top results to retrieve for each query.

        Returns:
            list | list[list]: The top results for the query. If query is a list, the top
                results for each query in the list are returned. Each list will contain the
                raw elements yielded by the index. This way, users can program against the
                results they expect to get from e.g. chromadb or ragatouille.
        """
        pass


class ChromaIndex(BaseIndex):
    def __init__(self, index: Collection):
        assert isinstance(index, Collection), "ChromaIndex input must be a `chromadb.api.models.Collection.Collection`"
        super().__init__(index)


# define type for PZIndex
PZIndex = ChromaIndex



from __future__ import annotations

import time
from typing import Callable

from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.fields import Field
from palimpzest.query.operators.physical import PhysicalOperator


class MapOp(PhysicalOperator):
    def __init__(self, udf: Callable | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.udf = udf

    def __str__(self):
        op = super().__str__()
        op += f"    UDF: {self.udf.__name__}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {"udf": self.udf, **id_params}

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {"udf": self.udf, **op_params}

        return op_params

    def _create_record_set(
        self,
        record: DataRecord,
        generation_stats: GenerationStats,
        total_time: float,
    ) -> DataRecordSet:
        """
        Given an input DataRecord and a determination of whether it passed the filter or not,
        construct the resulting RecordSet.
        """
        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=record.id,
            record_parent_id=record.parent_id,
            record_source_idx=record.source_idx,
            record_state=record.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=total_time,
            cost_per_record=0.0,
            fn_call_duration_secs=generation_stats.fn_call_duration_secs,
            answer=record.to_dict(include_bytes=False),
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([record], [record_op_stats])

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the Map operation. These estimates assume that the map UDF
        (1) has no cost and (2) has perfect quality.
        """
        # estimate 1 ms single-threaded execution for udf function
        time_per_record = 0.001

        # assume filter fn has perfect quality
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=time_per_record,
            cost_per_record=0.0,
            quality=1.0,
        )

    def map(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # apply UDF to input record
        start_time = time.time()
        field_answers = {}
        try:
            # execute the UDF function
            field_answers = self.udf(candidate.to_dict())

            # answer should be a dictionary
            assert isinstance(field_answers, dict), (
                "UDF must return a dictionary mapping each input field to its value for map operations"
            )

            if self.verbose:
                print(f"{self.udf.__name__}")

        except Exception as e:
            print(f"Error invoking user-defined function for map: {e}")
            raise e

        # create generation stats object containing the time spent executing the UDF function
        generation_stats = GenerationStats(fn_call_duration_secs=time.time() - start_time)

        return field_answers, generation_stats


    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        """
        This method converts an input DataRecord into an output DataRecordSet. The output DataRecordSet contains the
        DataRecord(s) output by the operator's convert() method and their corresponding RecordOpStats objects.
        Some subclasses may override this __call__method to implement their own custom logic.
        """
        start_time = time.time()

        # execute the map operation
        field_answers: dict[str, list]
        fields = {field: field_type for field, field_type in self.output_schema.field_map().items()}
        field_answers, generation_stats = self.map(candidate=candidate, fields=fields)
        assert all([field in field_answers for field in fields]), "Not all fields are present in output of map!"

        # construct DataRecord from field_answers
        dr = DataRecord.from_parent(schema=self.output_schema, parent_record=candidate)
        for field_name, field_value in field_answers.items():
            dr[field_name] = field_value

        # construct and return DataRecordSet
        record_set = self._create_record_set(
            record=dr,
            generation_stats=generation_stats,
            total_time=time.time() - start_time,
        )

        return record_set



import logging

from palimpzest.core.data.dataclasses import PlanStats
from palimpzest.core.elements.records import DataRecord
from palimpzest.query.execution.execution_strategy import ExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)

class SequentialSingleThreadExecutionStrategy(ExecutionStrategy):
    """
    A single-threaded execution strategy that processes operators sequentially.
    
    This strategy processes all records through one operator completely before moving to the next operator
    in the execution plan. For example, if we have operators A -> B -> C and records [1,2,3]:
    1. First processes records [1,2,3] through operator A
    2. Then takes A's output and processes all of it through operator B
    3. Finally processes all of B's output through operator C
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = 1

    def _execute_plan(self, plan: PhysicalPlan, input_queues: dict[str, list], plan_stats: PlanStats) -> tuple[list[DataRecord], PlanStats]:
        # execute the plan one operator at a time
        output_records = []
        for op_idx, operator in enumerate(plan.operators):
            # if we've filtered out all records, terminate early
            full_op_id = operator.get_full_op_id()
            num_inputs = len(input_queues[full_op_id])
            if num_inputs == 0:
                break

            # begin to process this operator
            records, record_op_stats = [], []
            logger.info(f"Processing operator {operator.op_name()} ({full_op_id})")

            # if this operator is an aggregate, process all the records in the input_queue
            if isinstance(operator, AggregateOp):
                record_set = operator(candidates=input_queues[full_op_id])
                records = record_set.data_records
                record_op_stats = record_set.record_op_stats
                num_outputs = sum(record.passed_operator for record in records)

                # update the progress manager
                self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

            # otherwise, process the records in the input queue for this operator one at a time
            else:
                for input_record in input_queues[full_op_id]:
                    record_set = operator(input_record)
                    records.extend(record_set.data_records)
                    record_op_stats.extend(record_set.record_op_stats)
                    num_outputs = sum(record.passed_operator for record in record_set.data_records)

                    # update the progress manager
                    self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                    # finish early if this is a limit
                    if isinstance(operator, LimitScanOp) and len(records) == operator.limit:
                        break

            # update plan stats
            plan_stats.add_record_op_stats(record_op_stats)

            # add records to the cache
            self._add_records_to_cache(operator.target_cache_id, records)

            # update next input_queue (if it exists)
            output_records = [record for record in records if record.passed_operator]            
            if op_idx + 1 < len(plan.operators):
                next_full_op_id = plan.operators[op_idx + 1].get_full_op_id()
                input_queues[next_full_op_id] = output_records

            logger.info(f"Finished processing operator {operator.op_name()} ({operator.get_full_op_id()}), and generated {len(records)} records")

        # close the cache
        self._close_cache([op.target_cache_id for op in plan.operators])

        # finalize plan stats
        plan_stats.finish()

        return output_records, plan_stats

    def execute_plan(self, plan: PhysicalPlan) -> tuple[list[DataRecord], PlanStats]:
        """Initialize the stats and execute the plan."""
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert isinstance(plan.operators[0], ScanPhysicalOp), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues for each operation
        input_queues = self._create_input_queues(plan)

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, num_samples=self.num_samples, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats


class PipelinedSingleThreadExecutionStrategy(ExecutionStrategy):
    """
    A single-threaded execution strategy that processes records through a pipeline of operators.
    
    This strategy implements a pipelined execution model where each record flows through
    the entire operator chain before the next record is processed.

    Example Flow:
    For operators A -> B -> C and records [1,2,3]:
    1. Record 1: A -> B -> C
    2. Record 2: A -> B -> C
    3. Record 3: A -> B -> C
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = 1

    def _any_queue_not_empty(self, queues: dict[str, list]) -> bool:
        """Helper function to check if any queue is not empty."""
        return any(len(queue) > 0 for queue in queues.values())

    def _upstream_ops_finished(self, plan: PhysicalPlan, op_idx: int, input_queues: dict[str, list]) -> bool:
        """Helper function to check if all upstream operators have finished processing their inputs."""
        for upstream_op_idx in range(op_idx):
            upstream_full_op_id = plan.operators[upstream_op_idx].get_full_op_id()
            if len(input_queues[upstream_full_op_id]) > 0:
                return False

        return True

    def _execute_plan(self, plan: PhysicalPlan, input_queues: dict[str, list], plan_stats: PlanStats) -> tuple[list[DataRecord], PlanStats]:
        # execute the plan until either:
        # 1. all records have been processed, or
        # 2. the final limit operation has completed (we break out of the loop if this happens)
        final_output_records = []
        while self._any_queue_not_empty(input_queues):
            for op_idx, operator in enumerate(plan.operators):
                # if this operator does not have enough inputs to execute, then skip it
                full_op_id = operator.get_full_op_id()
                num_inputs = len(input_queues[full_op_id])
                agg_op_not_ready = isinstance(operator, AggregateOp) and not self._upstream_ops_finished(plan, op_idx, input_queues)
                if num_inputs == 0 or agg_op_not_ready:
                    continue

                # create empty lists for records and execution stats generated by executing this operator on its next input(s)
                records, record_op_stats = [], []

                # if the next operator is an aggregate, process all the records in the input_queue
                if isinstance(operator, AggregateOp):
                    input_records = [input_queues[full_op_id].pop(0) for _ in range(num_inputs)]
                    record_set = operator(candidates=input_records)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    num_outputs = sum(record.passed_operator for record in records)

                    # update the progress manager
                    self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                # otherwise, process the next record in the input queue for this operator
                else:
                    input_record = input_queues[full_op_id].pop(0)
                    record_set = operator(input_record)
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    num_outputs = sum(record.passed_operator for record in records)

                    # update the progress manager
                    self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

                # update plan stats
                plan_stats.add_record_op_stats(record_op_stats)

                # add records to the cache
                self._add_records_to_cache(operator.target_cache_id, records)

                # update next input_queue or final_output_records
                output_records = [record for record in records if record.passed_operator]            
                if op_idx + 1 < len(plan.operators):
                    next_full_op_id = plan.operators[op_idx + 1].get_full_op_id()
                    input_queues[next_full_op_id].extend(output_records)
                else:
                    final_output_records.extend(output_records)

                logger.info(f"Finished processing operator {operator.op_name()} ({operator.get_full_op_id()}) on {num_inputs} records")

            # break out of loop if the final operator is a LimitScanOp and we've reached its limit
            if isinstance(plan.operators[-1], LimitScanOp) and len(final_output_records) == plan.operators[-1].limit:
                break

        # close the cache
        self._close_cache([op.target_cache_id for op in plan.operators])

        # finalize plan stats
        plan_stats.finish()

        return final_output_records, plan_stats

    def execute_plan(self, plan: PhysicalPlan):
        """Initialize the stats and execute the plan."""
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert isinstance(plan.operators[0], ScanPhysicalOp), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues for each operation
        input_queues = self._create_input_queues(plan)        

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, self.num_samples, self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats



import modal

app = modal.App("palimpzest.tools")
pip_packs = [
    "papermage",
    "tqdm",
    "transformers",
    "pdf2image",
    "pdfplumber==0.7.4",
    "requests",
    "numpy>=1.23.2",
    "scipy>=1.9.0",
    "pandas<2",
    "ncls==0.0.68",
    "necessary>=0.3.2",
    "grobid-client-python==0.0.5",
    "charset-normalizer",
    "torch>=1.10.0",
    "smashed",
    "layoutparser",
    "pysbd",
    "decontext",
    "vila",
]

pdf_processing_image = (
    modal.Image.debian_slim(python_version="3.11")
    .apt_install(["ffmpeg", "pkg-config", "libpoppler-cpp-dev", "poppler-utils"])
    .pip_install(["torch==2.1.1", "pkgconfig", "python-poppler"] + pip_packs)
)


@app.function(image=pdf_processing_image)
def process_papermage_pdf(pdf_bytes_docs: list[bytes]):
    """Process a PDF file and return the text contents."""
    import json
    import os

    from papermage.recipes import CoreRecipe

    os.makedirs("/tmp", exist_ok=True)
    results = []
    for pdf_bytes in pdf_bytes_docs:
        recipe = CoreRecipe()

        with open("/tmp/papermage.pdf", "wb") as file:
            file.write(pdf_bytes)

        doc = recipe.run("/tmp/papermage.pdf")

        os.remove("/tmp/papermage.pdf")

        results.append(json.dumps(doc.to_json()))

    return results


@app.local_entrypoint()
def main():
    import json

    from papermage import Document

    with open("test.pdf", "rb") as file:
        pdf_bytes_1 = file.read()

    results = process_papermage_pdf.local([pdf_bytes_1])
    for idx, r in enumerate(results):
        docdict = json.loads(r)
        doc = Document.from_json(docdict)
        print(idx, doc)
        for p in doc.pages:
            print(p)



from palimpzest.query.optimizer.rules import (
    AggregateRule as _AggregateRule,
)
from palimpzest.query.optimizer.rules import (
    BasicSubstitutionRule as _BasicSubstitutionRule,
)
from palimpzest.query.optimizer.rules import (
    CodeSynthesisConvertRule as _CodeSynthesisConvertRule,
)
from palimpzest.query.optimizer.rules import (
    CodeSynthesisConvertSingleRule as _CodeSynthesisConvertSingleRule,
)
from palimpzest.query.optimizer.rules import (
    CriticAndRefineConvertRule as _CriticAndRefineConvertRule,
)
from palimpzest.query.optimizer.rules import (
    ImplementationRule as _ImplementationRule,
)
from palimpzest.query.optimizer.rules import (
    LLMConvertBondedRule as _LLMConvertBondedRule,
)
from palimpzest.query.optimizer.rules import (
    LLMFilterRule as _LLMFilterRule,
)
from palimpzest.query.optimizer.rules import (
    MixtureOfAgentsConvertRule as _MixtureOfAgentsConvertRule,
)
from palimpzest.query.optimizer.rules import (
    NonLLMConvertRule as _NonLLMConvertRule,
)
from palimpzest.query.optimizer.rules import (
    NonLLMFilterRule as _NonLLMFilterRule,
)
from palimpzest.query.optimizer.rules import (
    PushDownFilter as _PushDownFilter,
)
from palimpzest.query.optimizer.rules import (
    RAGConvertRule as _RAGConvertRule,
)
from palimpzest.query.optimizer.rules import (
    RetrieveRule as _RetrieveRule,
)
from palimpzest.query.optimizer.rules import (
    Rule as _Rule,
)
from palimpzest.query.optimizer.rules import (
    SplitConvertRule as _SplitConvertRule,
)
from palimpzest.query.optimizer.rules import (
    TransformationRule as _TransformationRule,
)

ALL_RULES = [
    _AggregateRule,
    _BasicSubstitutionRule,
    _CodeSynthesisConvertRule,
    _CodeSynthesisConvertSingleRule,
    _CriticAndRefineConvertRule,
    _ImplementationRule,
    _LLMConvertBondedRule,
    _LLMFilterRule,
    _MixtureOfAgentsConvertRule,
    _NonLLMConvertRule,
    _NonLLMFilterRule,
    _PushDownFilter,
    _RAGConvertRule,
    _RetrieveRule,
    _Rule,
    _SplitConvertRule,
    _TransformationRule,
]

IMPLEMENTATION_RULES = [
    rule
    for rule in ALL_RULES
    if issubclass(rule, _ImplementationRule)
    and rule not in [_CodeSynthesisConvertRule, _ImplementationRule]
]

TRANSFORMATION_RULES = [
    rule for rule in ALL_RULES if issubclass(rule, _TransformationRule) and rule not in [_TransformationRule]
]

__all__ = [
    "ALL_RULES",
    "IMPLEMENTATION_RULES",
    "TRANSFORMATION_RULES",
]



from __future__ import annotations

import logging
from abc import ABC, abstractmethod

from palimpzest.policy import Policy
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan

logger = logging.getLogger(__name__)


class OptimizationStrategy(ABC):
    @abstractmethod
    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan] | list[SentinelPlan]:
        """Strategy decides how to search through the groups for optimal plan(s)"""
        pass

    def normalize_final_plans(self, plans: list[PhysicalPlan]) -> list[PhysicalPlan]:
        """
        For each plan in `plans`, this function enforces that the input schema of every
        operator is the output schema of the previous operator in the plan.

        Args:
            plans list[PhysicalPlan]: list of physical plans to normalize

        Returns:
            list[PhysicalPlan]: list of normalized physical plans
        """
        normalized_plans = []
        for plan in plans:
            normalized_ops = []
            for idx, op in enumerate(plan.operators):
                op_copy = op.copy()
                if idx == 0:
                    normalized_ops.append(op_copy)
                else:
                    op_copy.input_schema = plan.operators[-1].output_schema
                    normalized_ops.append(op_copy)
            normalized_plans.append(PhysicalPlan(operators=normalized_ops, plan_cost=plan.plan_cost))

        return normalized_plans


class GreedyStrategy(OptimizationStrategy):
    def _get_greedy_physical_plan(self, groups: dict, group_id: int) -> PhysicalPlan:
        """
        Return the best plan with respect to the user provided policy.
        """
        # get the best physical expression for this group
        best_phys_expr = groups[group_id].best_physical_expression

        # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
        # create and return the physical plan
        if len(best_phys_expr.input_group_ids) == 0:
            return PhysicalPlan(operators=[best_phys_expr.operator], plan_cost=best_phys_expr.plan_cost)

        # get the best physical plan(s) for this group's inputs
        input_group_id = best_phys_expr.input_group_ids[0] # TODO: need to handle joins
        input_best_phys_plan = self._get_greedy_physical_plan(groups, input_group_id)

        # add this operator to best physical plan and return
        return PhysicalPlan.from_ops_and_sub_plan([best_phys_expr.operator], input_best_phys_plan, best_phys_expr.plan_cost)

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan]:
        logger.info(f"Getting greedy optimal plans for final group id: {final_group_id}")
        plans = [self._get_greedy_physical_plan(groups, final_group_id)]
        logger.info(f"Greedy optimal plans: {plans}")
        logger.info(f"Done getting greedy optimal plans for final group id: {final_group_id}")

        return plans


class ParetoStrategy(OptimizationStrategy):
    def _get_candidate_pareto_physical_plans(self, groups: dict, group_id: int, policy: Policy) -> list[PhysicalPlan]:
        """
        Return a list of plans which will contain all of the pareto optimal plans (and some additional
        plans which may not be pareto optimal).

        TODO: can we cache group_id --> final_pareto_optimal_plans to avoid re-computing upstream
        groups' pareto-optimal plans for each expression?
        """
        # get the pareto optimal physical expressions for this group
        pareto_optimal_phys_exprs = groups[group_id].pareto_optimal_physical_expressions

        # construct list of pareto optimal plans
        pareto_optimal_plans = []
        for phys_expr in pareto_optimal_phys_exprs:
            # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
            # create and return the physical plan
            if len(phys_expr.input_group_ids) == 0:
                for plan_cost, _ in phys_expr.pareto_optimal_plan_costs:
                    plan = PhysicalPlan(operators=[phys_expr.operator], plan_cost=plan_cost)
                    pareto_optimal_plans.append(plan)

            # otherwise, get the pareto optimal physical plan(s) for this group's inputs
            else:
                # get the pareto optimal physical plan(s) for this group's inputs
                input_group_id = phys_expr.input_group_ids[0] # TODO: need to handle joins
                pareto_optimal_phys_subplans = self._get_candidate_pareto_physical_plans(groups, input_group_id, policy)

                # iterate over the input subplans and find the one(s) which combine with this physical expression
                # to make a pareto-optimal plan
                for plan_cost, input_plan_cost in phys_expr.pareto_optimal_plan_costs:
                    for subplan in pareto_optimal_phys_subplans:
                        if (
                            subplan.plan_cost.cost == input_plan_cost.cost
                            and subplan.plan_cost.time == input_plan_cost.time
                            and subplan.plan_cost.quality == input_plan_cost.quality
                        ):
                            # TODO: The plan_cost gets summed with subplan.plan_cost;
                            #       am I defining expression.best_plan_cost to be the cost of that operator,
                            #       and expression.pareto_optimal_plan_costs to be the cost(s) of the subplan including that operator?
                            #       i.e. are my definitions inconsistent?
                            plan = PhysicalPlan.from_ops_and_sub_plan([phys_expr.operator], subplan, plan_cost)
                            pareto_optimal_plans.append(plan)

        return pareto_optimal_plans

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[PhysicalPlan]:
        logger.info(f"Getting pareto optimal plans for final group id: {final_group_id}")
        # compute all of the pareto optimal physical plans
        plans = self._get_candidate_pareto_physical_plans(groups, final_group_id, policy)

        # adjust plans' plan_cost.quality to reflect only the quality of the final operator
        if use_final_op_quality:
            for plan in plans:
                plan.plan_cost.quality = plan.plan_cost.op_estimates.quality

        # filter pareto optimal plans for ones which satisfy policy constraint (if at least one of them does)
        if any([policy.constraint(plan.plan_cost) for plan in plans]):
            plans = [plan for plan in plans if policy.constraint(plan.plan_cost)]

        # select the plan which is best for the given policy
        optimal_plan, plans = plans[0], plans[1:]
        for plan in plans:
            optimal_plan = optimal_plan if policy.choose(optimal_plan.plan_cost, plan.plan_cost) else plan

        plans = [optimal_plan]
        logger.info(f"Pareto optimal plans: {plans}")
        logger.info(f"Done getting pareto optimal plans for final group id: {final_group_id}")
        return plans
    

class SentinelStrategy(OptimizationStrategy):
    def _get_sentinel_plan(self, groups: dict, group_id: int) -> SentinelPlan:
        """
        Create and return a SentinelPlan object.
        """
        # get all the physical expressions for this group
        phys_exprs = groups[group_id].physical_expressions
        phys_op_set = [expr.operator for expr in phys_exprs]

        # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
        # create and return the physical plan
        best_phys_expr = groups[group_id].best_physical_expression
        if len(best_phys_expr.input_group_ids) == 0:
            return SentinelPlan(operator_sets=[phys_op_set])

        # TODO: need to handle joins
        # get the best physical plan(s) for this group's inputs
        best_phys_subplan = SentinelPlan(operator_sets=[])
        for input_group_id in best_phys_expr.input_group_ids:
            input_best_phys_plan = self._get_sentinel_plan(groups, input_group_id)
            best_phys_subplan = SentinelPlan.from_ops_and_sub_plan(best_phys_subplan.operator_sets, input_best_phys_plan)

        # add this operator set to best physical plan and return
        return SentinelPlan.from_ops_and_sub_plan([phys_op_set], best_phys_subplan)

    def get_optimal_plans(self, groups: dict, final_group_id: int, policy: Policy, use_final_op_quality: bool) -> list[SentinelPlan]:
        logger.info(f"Getting sentinel optimal plans for final group id: {final_group_id}")
        plans = [self._get_sentinel_plan(groups, final_group_id)]
        logger.info(f"Sentinel optimal plans: {plans}")
        logger.info(f"Done getting sentinel optimal plans for final group id: {final_group_id}")
        return plans


class NoOptimizationStrategy(GreedyStrategy):
    """
    NoOptimizationStrategy is used to intentionally construct a PhysicalPlan without applying any
    logical transformations or optimizations. It uses the same get_optimal_plans logic as the
    GreedyOptimizationStrategy.
    """



### CODE SYNTHESIS PROMPTS ###
EXAMPLE_PROMPT = """Example{idx}:
Example Input
-------------
{example_inputs}

Example Output
--------------
{example_output}
"""

CODEGEN_PROMPT = """You are a helpful programming assistant and an expert {language} programmer. Implement the {language} function `{api}` that extracts `{output}` ({output_desc}) from given inputs:
{inputs_desc}
{examples_desc}
Notice that the evaluation will severely punish incorrect outputs. Thus, when the function is uncertain, it should return `None` to abstain instead of returning an incorrect guess.
{advice}
Return the implementation only."""

ADVICEGEN_PROMPT = """You are a helpful programming assistant and an expert {language} programmer. Your job is to provide programming ideas to help me write {language} programs.
For example, if I want to complete a task: "extract the salary number (in USD) from a given employee's document", you can provide me with {n} different ways to do it like:
Idea 1: Use regular expressions to extract the salary number: a number with a dollar sign in front of it. For example, $100,000.
Idea 2: Find the table entry with the salary number.
Idea 3: Use a pre-trained NLP model to extract the salary number.
# 
Now, consider the following {language} programming task that extracts `{output}` ({output_desc}) from given inputs:
{examples_desc}
Please provide me with {n} different ideas to complete this task. Return the ideas only, following the format above.
"""



from __future__ import annotations

import logging
from copy import deepcopy

from palimpzest.constants import Model
from palimpzest.core.data.datareaders import DataReader
from palimpzest.core.lib.fields import Field
from palimpzest.policy import Policy
from palimpzest.query.operators.logical import (
    Aggregate,
    BaseScan,
    ConvertScan,
    FilteredScan,
    GroupByAggregate,
    LimitScan,
    LogicalOperator,
    MapScan,
    Project,
    RetrieveScan,
)
from palimpzest.query.optimizer import (
    IMPLEMENTATION_RULES,
    TRANSFORMATION_RULES,
)
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.query.optimizer.primitives import Group, LogicalExpression
from palimpzest.query.optimizer.rules import (
    CodeSynthesisConvertRule,
    CriticAndRefineConvertRule,
    LLMConvertBondedRule,
    MixtureOfAgentsConvertRule,
    RAGConvertRule,
    SplitConvertRule,
)
from palimpzest.query.optimizer.tasks import (
    ApplyRule,
    ExpandGroup,
    OptimizeGroup,
    OptimizeLogicalExpression,
    OptimizePhysicalExpression,
)
from palimpzest.sets import Dataset, Set
from palimpzest.utils.hash_helpers import hash_for_serialized_dict
from palimpzest.utils.model_helpers import get_champion_model, get_code_champion_model, get_fallback_model

logger = logging.getLogger(__name__)


def get_node_uid(node: Dataset | DataReader) -> str:
    """Helper function to compute the universal identifier for a node in the query plan."""
    # NOTE: technically, hash_for_serialized_dict(node.serialize()) would be valid for both DataReader and Dataset;
    #       for the moment, I want to be explicit in Dataset about what constitutes a unique Dataset object, but
    #       in ther future we may be able to remove universal_identifier() from Dataset and just use this function
    return node.universal_identifier() if isinstance(node, Dataset) else hash_for_serialized_dict(node.serialize())


class Optimizer:
    """
    The optimizer is responsible for searching the space of possible physical plans
    for a user's initial (logical) plan and selecting the one which is closest to
    optimizing the user's policy objective.

    This optimizer is modeled after the Cascades framework for top-down query optimization:
    - Thesis describing Cascades implementation (Chapters 1-3):
      https://15721.courses.cs.cmu.edu/spring2023/papers/17-optimizer2/xu-columbia-thesis1998.pdf

    - Andy Pavlo lecture with walkthrough example: https://www.youtube.com/watch?v=PXS49-tFLcI

    - Original Paper: https://www.cse.iitb.ac.in/infolab/Data/Courses/CS632/2015/Papers/Cascades-graefe.pdf

    Notably, this optimization framework has served as the backbone of Microsoft SQL Server, CockroachDB,
    and a few other important DBMS systems.

    NOTE: the optimizer currently assumes that field names are unique across schemas; we do try to enforce
          this by rewriting field names underneath-the-hood to be "{schema_name}.{field_name}", but this still
          does not solve a situation in which -- for example -- a user uses the pz.URL schema twice in the same
          program. In order to address that situation, we will need to augment our renaming scheme.
    """

    def __init__(
        self,
        policy: Policy,
        cost_model: CostModel,
        available_models: list[Model],
        cache: bool = False,
        verbose: bool = False,
        allow_bonded_query: bool = True,
        allow_code_synth: bool = False,
        allow_rag_reduction: bool = False,
        allow_mixtures: bool = True,
        allow_critic: bool = False,
        allow_split_merge: bool = False,
        optimizer_strategy: OptimizationStrategyType = OptimizationStrategyType.PARETO,
        use_final_op_quality: bool = False, # TODO: make this func(plan) -> final_quality
        **kwargs,
    ):
        # store the policy
        self.policy = policy

        # store the cost model
        self.cost_model = cost_model

        # mapping from each group id to its Group object
        self.groups = {}

        # mapping from each expression to its Expression object
        self.expressions = {}

        # the stack of tasks to perform during optimization
        self.tasks_stack = []

        # the lists of implementation and transformation rules that the optimizer can apply
        self.implementation_rules = IMPLEMENTATION_RULES
        self.transformation_rules = TRANSFORMATION_RULES

        # get the strategy class associated with the optimizer strategy
        optimizer_strategy_cls = optimizer_strategy.value
        self.strategy = optimizer_strategy_cls()

        # remove transformation rules for optimization strategies which do not require them
        if optimizer_strategy.no_transformation():
            self.transformation_rules = []

        # if we are not performing optimization, set available models to be single model
        # and remove all optimizations (except for bonded queries)
        if optimizer_strategy == OptimizationStrategyType.NONE:
            self.allow_bonded_query = True
            self.allow_code_synth = False
            self.allow_rag_reduction = False
            self.allow_mixtures = False
            self.allow_critic = False
            self.allow_split_merge = False
            self.available_models = [available_models[0]]

        # store optimization hyperparameters
        self.cache = cache
        self.verbose = verbose
        self.available_models = available_models
        self.allow_bonded_query = allow_bonded_query
        self.allow_code_synth = allow_code_synth
        self.allow_rag_reduction = allow_rag_reduction
        self.allow_mixtures = allow_mixtures
        self.allow_critic = allow_critic
        self.allow_split_merge = allow_split_merge
        self.optimizer_strategy = optimizer_strategy
        self.use_final_op_quality = use_final_op_quality

        # prune implementation rules based on boolean flags
        if not self.allow_bonded_query:
            self.implementation_rules = [
                rule
                for rule in self.implementation_rules
                if rule not in [LLMConvertBondedRule]
            ]

        if not self.allow_code_synth:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, CodeSynthesisConvertRule)
            ]

        if not self.allow_rag_reduction:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, RAGConvertRule)
            ]

        if not self.allow_mixtures:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, MixtureOfAgentsConvertRule)
            ]

        if not self.allow_critic:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, CriticAndRefineConvertRule)
            ]

        if not self.allow_split_merge:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, SplitConvertRule)
            ]

        logger.info(f"Initialized Optimizer with verbose={self.verbose}")
        logger.debug(f"Initialized Optimizer with params: {self.__dict__}")

    def update_cost_model(self, cost_model: CostModel):
        self.cost_model = cost_model

    def get_physical_op_params(self):
        return {
            "verbose": self.verbose,
            "available_models": self.available_models,
            "champion_model": get_champion_model(self.available_models),
            "code_champion_model": get_code_champion_model(self.available_models),
            "fallback_model": get_fallback_model(self.available_models),
        }

    def deepcopy_clean(self):
        optimizer = Optimizer(
            policy=self.policy,
            cost_model=CostModel(),
            cache=self.cache,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_code_synth=self.allow_code_synth,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            allow_critic=self.allow_critic,
            allow_split_merge=self.allow_split_merge,
            optimizer_strategy=self.optimizer_strategy,
            use_final_op_quality=self.use_final_op_quality,
        )
        return optimizer

    def update_strategy(self, optimizer_strategy: OptimizationStrategyType):
        self.optimizer_strategy = optimizer_strategy
        optimizer_strategy_cls = optimizer_strategy.value
        self.strategy = optimizer_strategy_cls()

    def construct_group_tree(self, dataset_nodes: list[Set]) -> tuple[list[int], dict[str, Field], dict[str, set[str]]]:
        # get node, output_schema, and input_schema (if applicable)
        logger.debug(f"Constructing group tree for dataset_nodes: {dataset_nodes}")

        node = dataset_nodes[-1]
        output_schema = node.schema
        input_schema = dataset_nodes[-2].schema if len(dataset_nodes) > 1 else None

        ### convert node --> Group ###
        uid = get_node_uid(node)

        # create the op for the given node
        op: LogicalOperator | None = None

        # TODO: add cache scan when we add caching back to PZ
        # if self.cache:
        #     op = CacheScan(datareader=node, output_schema=output_schema)
        if isinstance(node, DataReader):
            op = BaseScan(datareader=node, output_schema=output_schema)
        elif node._filter is not None:
            op = FilteredScan(
                input_schema=input_schema,
                output_schema=output_schema,
                filter=node._filter,
                depends_on=node._depends_on,
                target_cache_id=uid,
            )
        elif node._group_by is not None:
            op = GroupByAggregate(
                input_schema=input_schema,
                output_schema=output_schema,
                group_by_sig=node._group_by,
                target_cache_id=uid,
            )
        elif node._agg_func is not None:
            op = Aggregate(
                input_schema=input_schema,
                output_schema=output_schema,
                agg_func=node._agg_func,
                target_cache_id=uid,
            )
        elif node._limit is not None:
            op = LimitScan(
                input_schema=input_schema,
                output_schema=output_schema,
                limit=node._limit,
                target_cache_id=uid,
            )
        elif node._project_cols is not None:
            op = Project(
                input_schema=input_schema,
                output_schema=output_schema,
                project_cols=node._project_cols,
                target_cache_id=uid,
            )
        elif node._index is not None:
            op = RetrieveScan(
                input_schema=input_schema,
                output_schema=output_schema,
                index=node._index,
                search_func=node._search_func,
                search_attr=node._search_attr,
                output_attrs=node._output_attrs,
                k=node._k,
                target_cache_id=uid,
            )
        elif output_schema != input_schema:
            op = ConvertScan(
                input_schema=input_schema,
                output_schema=output_schema,
                cardinality=node._cardinality,
                udf=node._udf,
                depends_on=node._depends_on,
                target_cache_id=uid,
            )
        elif output_schema == input_schema and node._udf is not None:
            op = MapScan(
                input_schema=input_schema,
                output_schema=output_schema,
                udf=node._udf,
                target_cache_id=uid,
            )
        # some legacy plans may have a useless convert; for now we simply skip it
        elif output_schema == input_schema:
            return self.construct_group_tree(dataset_nodes[:-1]) if len(dataset_nodes) > 1 else ([], {}, {})
        else:
            raise NotImplementedError(
                f"""No logical operator exists for the specified dataset construction.
                {input_schema}->{output_schema} {"with filter:'" + node._filter + "'" if node._filter is not None else ""}"""
            )

        # compute the input group ids and fields for this node
        input_group_ids, input_group_fields, input_group_properties = (
            self.construct_group_tree(dataset_nodes[:-1]) if len(dataset_nodes) > 1 else ([], {}, {})
        )

        # compute the fields added by this operation and all fields
        input_group_short_field_names = list(
            map(lambda full_field: full_field.split(".")[-1], input_group_fields.keys())
        )
        new_fields = {
            field_name: field
            for field_name, field in op.output_schema.field_map(unique=True, id=uid).items()
            if (field_name.split(".")[-1] not in input_group_short_field_names) or (node._udf is not None)
        }
        all_fields = {**input_group_fields, **new_fields}

        # compute the set of (short) field names this operation depends on
        depends_on_field_names = (
            {} if isinstance(node, DataReader) else {field_name.split(".")[-1] for field_name in node._depends_on}
        )

        # compute all properties including this operations'
        all_properties = deepcopy(input_group_properties)
        if isinstance(op, FilteredScan):
            # NOTE: we could use op.get_full_op_id() here, but storing filter strings makes
            #       debugging a bit easier as you can read which filters are in the Group
            op_filter_str = op.filter.get_filter_str()
            if "filters" in all_properties:
                all_properties["filters"].add(op_filter_str)
            else:
                all_properties["filters"] = set([op_filter_str])

        elif isinstance(op, LimitScan):
            op_limit_str = op.get_logical_op_id()
            if "limits" in all_properties:
                all_properties["limits"].add(op_limit_str)
            else:
                all_properties["limits"] = set([op_limit_str])

        elif isinstance(op, Project):
            op_project_str = op.get_logical_op_id()
            if "projects" in all_properties:
                all_properties["projects"].add(op_project_str)
            else:
                all_properties["projects"] = set([op_project_str])

        elif isinstance(op, MapScan):
            op_udf_str = op.udf.__name__
            if "udfs" in all_properties:
                all_properties["udfs"].add(op_udf_str)
            else:
                all_properties["udfs"] = set([op_udf_str])

        # construct the logical expression and group
        logical_expression = LogicalExpression(
            operator=op,
            input_group_ids=input_group_ids,
            input_fields=input_group_fields,
            depends_on_field_names=depends_on_field_names,
            generated_fields=new_fields,
            group_id=None,
        )
        group = Group(
            logical_expressions=[logical_expression],
            fields=all_fields,
            properties=all_properties,
        )
        logical_expression.set_group_id(group.group_id)

        # add the expression and group to the optimizer's expressions and groups and return
        self.expressions[logical_expression.get_expr_id()] = logical_expression
        self.groups[group.group_id] = group
        logger.debug(f"Constructed group tree for dataset_nodes: {dataset_nodes}")
        logger.debug(f"Group: {group.group_id}, {all_fields}, {all_properties}")

        return [group.group_id], all_fields, all_properties

    def convert_query_plan_to_group_tree(self, query_plan: Dataset) -> str:
        logger.debug(f"Converting query plan to group tree for query_plan: {query_plan}")
        # Obtain ordered list of datasets
        dataset_nodes: list[Dataset | DataReader] = []
        node = query_plan.copy()

        # NOTE: the very first node will be a DataReader; the rest will be Dataset
        while isinstance(node, Dataset):
            dataset_nodes.append(node)
            node = node._source
        dataset_nodes.append(node)
        dataset_nodes = list(reversed(dataset_nodes))

        # compute depends_on field for every node
        short_to_full_field_name = {}
        for node_idx, node in enumerate(dataset_nodes):
            # update mapping from short to full field names
            short_field_names = node.schema.field_names()
            full_field_names = node.schema.field_names(unique=True, id=get_node_uid(node))
            for short_field_name, full_field_name in zip(short_field_names, full_field_names):
                # set mapping automatically if this is a new field
                if short_field_name not in short_to_full_field_name or (
                    node_idx > 0 and dataset_nodes[node_idx - 1].schema != node.schema and node._udf is not None
                ):
                    short_to_full_field_name[short_field_name] = full_field_name

            # if the node is a data source, then skip
            if isinstance(node, DataReader):
                continue

            # If the node already has depends_on specified, then resolve each field name to a full (unique) field name
            if len(node._depends_on) > 0:
                node._depends_on = list(map(lambda field: short_to_full_field_name[field], node._depends_on))
                continue

            # otherwise, make the node depend on all upstream nodes
            node._depends_on = set()
            for upstream_node in dataset_nodes[:node_idx]:
                node._depends_on.update(upstream_node.schema.field_names(unique=True, id=get_node_uid(upstream_node)))
            node._depends_on = list(node._depends_on)

        # construct tree of groups
        final_group_id, _, _ = self.construct_group_tree(dataset_nodes)

        # check that final_group_id is a singleton
        assert len(final_group_id) == 1
        final_group_id = final_group_id[0]
        logger.debug(f"Converted query plan to group tree for query_plan: {query_plan}")
        logger.debug(f"Final group id: {final_group_id}")
        return final_group_id

    def heuristic_optimization(self, group_id: int) -> None:
        """
        Apply universally desirable transformations (e.g. filter/projection push-down).
        """
        pass

    def search_optimization_space(self, group_id: int) -> None:
        logger.debug(f"Searching optimization space for group_id: {group_id}")

        # begin the search for an optimal plan with a task to optimize the final group
        initial_task = OptimizeGroup(group_id)
        self.tasks_stack.append(initial_task)

        # TODO: conditionally stop when X number of tasks have been executed to limit exhaustive search
        while len(self.tasks_stack) > 0:
            task = self.tasks_stack.pop(-1)
            new_tasks = []
            if isinstance(task, (OptimizeGroup, ExpandGroup)):
                new_tasks = task.perform(self.groups)
            elif isinstance(task, OptimizeLogicalExpression):
                new_tasks = task.perform(self.transformation_rules, self.implementation_rules)
            elif isinstance(task, ApplyRule):
                context = {"costed_full_op_ids": self.cost_model.get_costed_full_op_ids()}
                new_tasks = task.perform(
                    self.groups, self.expressions, context=context, **self.get_physical_op_params()
                )
            elif isinstance(task, OptimizePhysicalExpression):
                context = {"optimizer_strategy": self.optimizer_strategy}
                new_tasks = task.perform(self.cost_model, self.groups, self.policy, context=context)

            self.tasks_stack.extend(new_tasks)

        logger.debug(f"Done searching optimization space for group_id: {group_id}")

    def optimize(self, query_plan: Dataset) -> list[PhysicalPlan]:
        """
        The optimize function takes in an initial query plan and searches the space of
        logical and physical plans in order to cost and produce a (near) optimal physical plan.
        """
        logger.info(f"Optimizing query plan: {query_plan}")
        # compute the initial group tree for the user plan
        final_group_id = self.convert_query_plan_to_group_tree(query_plan)

        # TODO
        # # do heuristic based pre-optimization
        # self.heuristic_optimization(final_group_id)

        # search the optimization space by applying logical and physical transformations to the initial group tree
        self.search_optimization_space(final_group_id)
        logger.info(f"Getting optimal plans for final group id: {final_group_id}")

        return self.strategy.get_optimal_plans(self.groups, final_group_id, self.policy, self.use_final_op_quality)



from __future__ import annotations

from palimpzest.core.lib.fields import Field
from palimpzest.query.operators.logical import LogicalOperator
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.optimizer.plan import PlanCost
from palimpzest.utils.hash_helpers import hash_for_id


class Expression:
    """
    An Expression (technically a "multi-expression") consists of either a logical operator
    (if it's a logical expression) or a physical operator (if it's a physical expression)
    and the group ids which are inputs to this expression
    """

    def __init__(
        self,
        operator: LogicalOperator | PhysicalOperator,
        input_group_ids: list[int],
        input_fields: dict[str, Field],
        depends_on_field_names: set[str],
        generated_fields: dict[str, Field],
        group_id: int | None = None,
    ):
        self.operator = operator
        self.input_group_ids = input_group_ids
        self.input_fields = input_fields
        self.depends_on_field_names = depends_on_field_names
        self.generated_fields = generated_fields
        self.group_id = group_id
        self.rules_applied = set()

        # NOTE: this will be the best possible plan cost achieved by this expression for some
        # greedy definition of "best"
        self.plan_cost: PlanCost | None = None

        # NOTE: this will be a list of tuples where each tuple has a (pareto-optimal) plan cost
        # and the input plan cost for which that pareto-optimal plan cost is attainable
        self.pareto_optimal_plan_costs: list[tuple[PlanCost, PlanCost]] | None = None

    def __eq__(self, other):
        return self.operator == other.operator and self.input_group_ids == other.input_group_ids

    def __str__(self):
        op_id = self.operator.get_logical_op_id() if isinstance(self.operator, LogicalOperator) else self.operator.get_full_op_id()
        return str(tuple(sorted(self.input_group_ids)) + (op_id, str(self.__class__.__name__)))

    def __hash__(self):
        hash_str = self.__str__()
        hash_id = int(hash_for_id(hash_str), 16)
        return hash_id

    def add_applied_rule(self, rule):
        self.rules_applied.add(rule.get_rule_id())

    def set_group_id(self, group_id: int) -> None:
        self.group_id = group_id

    def get_expr_id(self) -> int:
        return self.__hash__()


class LogicalExpression(Expression):
    pass


class PhysicalExpression(Expression):
    pass


class Group:
    """
    A group is a set of logically equivalent expressions (both logical (query trees) and physical (execution plans)).
    Represents the execution of an un-ordered set of logical operators.
    Maintains a set of logical multi-expressions and physical multi-expressions.
    """

    def __init__(self, logical_expressions: list[Expression], fields: dict[str, Field], properties: dict[str, set[str]]):
        self.logical_expressions = set(logical_expressions)
        self.physical_expressions = set()
        self.fields = fields
        self.explored = False
        self.best_physical_expression: PhysicalExpression | None = None
        self.pareto_optimal_physical_expressions: list[PhysicalExpression] | None = None
        self.optimized = False

        # properties of the Group which distinguish it from groups w/identical fields,
        # e.g. which filters, limits have been applied; is the output sorted, etc.
        self.properties = properties

        # compute the group id
        self.group_id = self.compute_group_id()

    def set_explored(self):
        self.explored = True

    def compute_group_id(self) -> int:
        # sort field names
        sorted_fields = sorted(self.fields.keys())

        # sort properties
        sorted_properties = []
        for key in sorted(self.properties.keys()):
            sorted_properties.extend(sorted(self.properties[key]))

        hash_str = str(tuple(sorted_fields + sorted_properties))
        hash_id = int(hash_for_id(hash_str), 16)
        return hash_id
